"""Azure DevOps アダプター。GitServiceAdapter の全メソッドを Azure DevOps REST API v7.1 で実装する。"""

from __future__ import annotations

import base64
import json as _json
from urllib.parse import quote, urlparse

import requests

from gfo.exceptions import GfoError, NotSupportedError
from gfo.http import paginate_top_skip

from .base import (
    Branch,
    BranchProtection,
    CheckRun,
    Comment,
    Commit,
    CommitStatus,
    CompareFile,
    CompareResult,
    DeployKey,
    GitServiceAdapter,
    Issue,
    IssueTemplate,
    Label,
    Milestone,
    Organization,
    Pipeline,
    PullRequest,
    PullRequestCommit,
    PullRequestFile,
    Release,
    Repository,
    Review,
    Tag,
    TimeEntry,
    TimelineEvent,
    Webhook,
)
from .registry import register


def _wiql_escape(value: str) -> str:
    """WIQL クエリ文字列内のシングルクォートをエスケープする。"""
    return value.replace("'", "''")


_PR_STATE_TO_API = {"open": "active", "closed": "abandoned", "merged": "completed"}
_PR_STATE_FROM_API = {"active": "open", "abandoned": "closed", "completed": "merged"}

_CLOSED_STATES = frozenset({"Closed", "Done", "Removed"})


def _add_refs_prefix(branch: str) -> str:
    if branch.startswith("refs/heads/"):
        return branch
    return f"refs/heads/{branch}"


def _strip_refs_prefix(ref: str) -> str:
    if ref.startswith("refs/heads/"):
        return ref[len("refs/heads/") :]
    return ref


@register("azure-devops")
class AzureDevOpsAdapter(GitServiceAdapter):
    service_name = "Azure DevOps"

    def __init__(
        self, client, owner: str, repo: str, *, organization: str, project_key: str, **kwargs
    ):
        super().__init__(client, owner, repo, **kwargs)
        self._org = organization
        self._project = project_key

    def _git_path(self) -> str:
        return f"/git/repositories/{quote(self._repo, safe='')}"

    def _wit_path(self) -> str:
        return "/wit"

    # --- 変換ヘルパー ---

    @staticmethod
    def _to_pull_request(data: dict) -> PullRequest:
        try:
            return PullRequest(
                number=data["pullRequestId"],
                title=data["title"],
                body=data.get("description"),
                state=_PR_STATE_FROM_API.get(data["status"], "open"),
                author=data["createdBy"]["uniqueName"],
                source_branch=_strip_refs_prefix(data["sourceRefName"]),
                target_branch=_strip_refs_prefix(data["targetRefName"]),
                draft=data.get("isDraft", False),
                url=(
                    f"{data['repository']['webUrl']}/pullrequest/{data['pullRequestId']}"
                    if (data.get("repository") or {}).get("webUrl")
                    else data.get("url", "")
                ),
                created_at=data["creationDate"],
                updated_at=data.get("closedDate"),
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    @staticmethod
    def _to_issue(data: dict) -> Issue:
        try:
            fields = data["fields"]
            raw_state = fields.get("System.State", "")
            state = "closed" if raw_state in _CLOSED_STATES else "open"

            assigned_to = fields.get("System.AssignedTo")
            unique_name = assigned_to.get("uniqueName") if assigned_to else None
            assignees = [unique_name] if unique_name else []

            raw_tags = fields.get("System.Tags", "")
            labels = [t.strip() for t in raw_tags.split(";") if t.strip()] if raw_tags else []

            return Issue(
                number=data["id"],
                title=fields["System.Title"],
                body=fields.get("System.Description"),
                state=state,
                author=(fields.get("System.CreatedBy") or {}).get("uniqueName", ""),
                assignees=assignees,
                labels=labels,
                url=data.get("url", ""),
                created_at=fields.get("System.CreatedDate", ""),
                updated_at=fields.get("System.ChangedDate"),
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    @staticmethod
    def _to_repository(data: dict, project: str = "") -> Repository:
        try:
            return Repository(
                name=data["name"],
                full_name=f"{project}/{data['name']}",
                description=(data.get("project") or {}).get("description"),
                private=True,
                default_branch=_strip_refs_prefix(data.get("defaultBranch") or ""),
                clone_url=data.get("remoteUrl", ""),
                url=data.get("webUrl", ""),
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    # --- PR ---

    def list_pull_requests(
        self,
        *,
        state: str = "open",
        limit: int = 30,
        author: str | None = None,
        label: str | None = None,
        assignee: str | None = None,
        search: str | None = None,
        base: str | None = None,
        head: str | None = None,
        draft: bool | None = None,
    ) -> list[PullRequest]:
        self._warn_unsupported_params("pr list", label=label, assignee=assignee, search=search)
        params: dict = {}
        if state != "all":
            params["searchCriteria.status"] = _PR_STATE_TO_API.get(state, "active")
        if base:
            params["searchCriteria.targetRefName"] = f"refs/heads/{base}"
        if head:
            params["searchCriteria.sourceRefName"] = f"refs/heads/{head}"
        results = paginate_top_skip(
            self._client,
            f"{self._git_path()}/pullrequests",
            params=params,
            limit=limit,
            result_key="value",
        )
        if author:
            results = [r for r in results if (r.get("createdBy") or {}).get("uniqueName") == author]
        if draft is not None:
            results = [r for r in results if r.get("isDraft", False) == draft]
        return [self._to_pull_request(r) for r in results]

    def create_pull_request(
        self,
        *,
        title: str,
        body: str = "",
        base: str,
        head: str,
        draft: bool = False,
        reviewers: list[str] | None = None,
        assignees: list[str] | None = None,
        labels: list[str] | None = None,
        milestone: str | None = None,
    ) -> PullRequest:
        self._warn_unsupported_params(
            "pull requests", assignees=assignees, labels=labels, milestone=milestone
        )
        payload = {
            "title": title,
            "description": body,
            "sourceRefName": _add_refs_prefix(head),
            "targetRefName": _add_refs_prefix(base),
            "isDraft": draft,
        }
        resp = self._client.post(f"{self._git_path()}/pullrequests", json=payload)
        pr = self._to_pull_request(resp.json())
        if reviewers:
            self.request_reviewers(pr.number, reviewers)
        return pr

    def get_pull_request(self, number: int) -> PullRequest:
        resp = self._client.get(f"{self._git_path()}/pullrequests/{number}")
        return self._to_pull_request(resp.json())

    def merge_pull_request(
        self,
        number: int,
        *,
        method: str = "merge",
        title: str | None = None,
        message: str | None = None,
    ) -> None:
        strategy_map = {
            "merge": "noFastForward",
            "squash": "squash",
            "rebase": "rebase",
        }
        strategy = strategy_map.get(method, "noFastForward")
        pr_resp = self._client.get(f"{self._git_path()}/pullrequests/{number}")
        pr_data = pr_resp.json()
        if not isinstance(pr_data, dict):
            raise GfoError(f"Unexpected API response from pullrequests endpoint: {type(pr_data)}")
        last_merge_commit = pr_data.get("lastMergeSourceCommit")
        if not last_merge_commit:
            raise GfoError(
                f"Cannot merge pull request #{number}: lastMergeSourceCommit not found. "
                "The pull request may have no commits or may be in an invalid state."
            )
        completion_options: dict = {"mergeStrategy": strategy}
        if title is not None or message is not None:
            parts = [p for p in (title, message) if p is not None]
            completion_options["mergeCommitMessage"] = "\n\n".join(parts)
        payload = {
            "status": "completed",
            "lastMergeSourceCommit": last_merge_commit,
            "completionOptions": completion_options,
        }
        self._client.patch(f"{self._git_path()}/pullrequests/{number}", json=payload)

    def close_pull_request(self, number: int) -> None:
        self._client.patch(
            f"{self._git_path()}/pullrequests/{number}",
            json={"status": "abandoned"},
        )

    def reopen_pull_request(self, number: int) -> None:
        self._client.patch(
            f"{self._git_path()}/pullrequests/{number}",
            json={"status": "active"},
        )

    def get_pr_checkout_refspec(self, number: int, *, pr: PullRequest | None = None) -> str:
        return f"refs/pull/{number}/head"

    def get_pull_request_diff(self, number: int) -> str:
        raise NotSupportedError(self.service_name, "pr diff")

    def list_pull_request_checks(self, number: int) -> list[CheckRun]:
        _STATE_MAP = {
            "succeeded": "success",
            "failed": "failure",
            "pending": "pending",
            "notApplicable": "success",
            "notSet": "pending",
            "error": "failure",
        }
        results = paginate_top_skip(
            self._client,
            f"{self._git_path()}/pullrequests/{number}/statuses",
            result_key="value",
        )
        out: list[CheckRun] = []
        for s in results:
            genre = s.get("context", {}).get("genre", "")
            name = s.get("context", {}).get("name", "")
            full_name = f"{genre}/{name}" if genre else name
            state = _STATE_MAP.get(s.get("state", ""), "pending")
            out.append(
                CheckRun(
                    name=full_name,
                    status=state,
                    conclusion="",
                    url=s.get("targetUrl", ""),
                    started_at=s.get("creationDate", ""),
                )
            )
        return out

    def list_pull_request_files(self, number: int) -> list[PullRequestFile]:
        _CHANGE_MAP = {
            "add": "added",
            "edit": "modified",
            "delete": "deleted",
            "rename": "renamed",
        }
        # 最新のイテレーションを取得
        iters = self._client.get(
            f"{self._git_path()}/pullrequests/{number}/iterations",
        ).json()
        iterations = iters.get("value", [])
        if not iterations:
            return []
        last_id = iterations[-1]["id"]
        resp = self._client.get(
            f"{self._git_path()}/pullrequests/{number}/iterations/{last_id}/changes",
        ).json()
        out: list[PullRequestFile] = []
        for entry in resp.get("changeEntries", []):
            item = entry.get("item", {})
            change_type = entry.get("changeType", "edit")
            # changeType はカンマ区切りで複合値になる場合がある（例: "rename, edit"）
            primary = change_type.split(",")[0].strip().lower()
            out.append(
                PullRequestFile(
                    filename=item.get("path", ""),
                    status=_CHANGE_MAP.get(primary, "modified"),
                    additions=0,
                    deletions=0,
                )
            )
        return out

    def list_pull_request_commits(self, number: int) -> list[PullRequestCommit]:
        results = paginate_top_skip(
            self._client,
            f"{self._git_path()}/pullrequests/{number}/commits",
            result_key="value",
        )
        return [
            PullRequestCommit(
                sha=c.get("commitId", ""),
                message=c.get("comment", ""),
                author=c.get("author", {}).get("name", ""),
                created_at=c.get("author", {}).get("date", ""),
            )
            for c in results
        ]

    def list_requested_reviewers(self, number: int) -> list[str]:
        resp = self._client.get(
            f"{self._git_path()}/pullrequests/{number}/reviewers",
        ).json()
        return [r.get("displayName", "") for r in resp.get("value", [])]

    def request_reviewers(self, number: int, reviewers: list[str]) -> None:
        for reviewer in reviewers:
            self._client.put(
                f"{self._git_path()}/pullrequests/{number}/reviewers/{quote(reviewer, safe='')}",
                json={"vote": 0},
            )

    def remove_reviewers(self, number: int, reviewers: list[str]) -> None:
        for reviewer in reviewers:
            self._client.delete(
                f"{self._git_path()}/pullrequests/{number}/reviewers/{quote(reviewer, safe='')}",
            )

    def enable_auto_merge(self, number: int, *, merge_method: str | None = None) -> None:
        _MERGE_STRATEGY = {
            "merge": "noFastForward",
            "squash": "squash",
            "rebase": "rebaseMerge",
        }
        pr_resp = self._client.get(
            f"{self._git_path()}/pullrequests/{number}",
        ).json()
        created_by_id = pr_resp.get("createdBy", {}).get("id", "")
        strategy = _MERGE_STRATEGY.get(merge_method or "", "noFastForward")
        self._client.patch(
            f"{self._git_path()}/pullrequests/{number}",
            json={
                "autoCompleteSetBy": {"id": created_by_id},
                "completionOptions": {"mergeStrategy": strategy},
            },
        )

    def mark_pull_request_ready(self, number: int) -> None:
        self._client.patch(
            f"{self._git_path()}/pullrequests/{number}",
            json={"isDraft": False},
        )

    def dismiss_review(self, number: int, review_id: int, *, message: str = "") -> None:
        self._client.put(
            f"{self._git_path()}/pullrequests/{number}/reviewers/{review_id}",
            json={"vote": 0},
        )

    # --- Issue (Work Item) ---

    def list_issues(
        self,
        *,
        state: str = "open",
        assignee: str | None = None,
        label: str | None = None,
        limit: int = 30,
        author: str | None = None,
        milestone: str | None = None,
        search: str | None = None,
    ) -> list[Issue]:
        self._warn_unsupported_params(
            "issue list", author=author, milestone=milestone, search=search
        )
        conditions = [f"[System.TeamProject] = '{_wiql_escape(self._project)}'"]
        if state == "open":
            conditions.append("[System.State] NOT IN ('Closed', 'Done', 'Removed')")
        elif state == "closed":
            conditions.append("[System.State] IN ('Closed', 'Done', 'Removed')")
        if assignee:
            conditions.append(f"[System.AssignedTo] = '{_wiql_escape(assignee)}'")
        if label:
            conditions.append(f"[System.Tags] CONTAINS '{_wiql_escape(label)}'")

        wiql = "SELECT [System.Id] FROM WorkItems WHERE " + " AND ".join(conditions)  # nosec B608 - conditions built from internal values only, no user input
        wiql_params = {"$top": limit} if limit > 0 else {"$top": 20000}
        wiql_resp = self._client.post(
            f"{self._wit_path()}/wiql",
            json={"query": wiql},
            params=wiql_params,
        )
        wiql_body = wiql_resp.json()
        if not isinstance(wiql_body, dict):
            raise GfoError(f"Unexpected API response from WIQL endpoint: {type(wiql_body)}")
        try:
            ids = [wi["id"] for wi in wiql_body.get("workItems", [])]
        except (KeyError, TypeError) as e:
            raise GfoError(f"Unexpected API response from WIQL endpoint: {e}") from e
        if not ids:
            return []

        results: list[Issue] = []
        for i in range(0, len(ids), 200):
            # limit > 0 の場合、WIQL が $top で件数を制限済みだが念のため早期終了
            if limit > 0 and len(results) >= limit:
                break
            batch_ids = ids[i : i + 200]
            ids_str = ",".join(str(x) for x in batch_ids)
            resp = self._client.get(
                f"{self._wit_path()}/workitems",
                params={"ids": ids_str, "$expand": "None"},
            )
            batch_body = resp.json()
            if not isinstance(batch_body, dict):
                raise GfoError(
                    f"Unexpected API response from workitems endpoint: {type(batch_body)}"
                )
            for item in batch_body.get("value", []):
                results.append(self._to_issue(item))

        return results[:limit] if limit > 0 else results

    def create_issue(
        self,
        *,
        title: str,
        body: str = "",
        assignee: str | None = None,
        label: str | None = None,
        milestone: str | None = None,
        work_item_type: str = "Task",
        **kwargs,
    ) -> Issue:
        self._warn_unsupported_params("issue create", milestone=milestone)
        patch_ops = [
            {"op": "add", "path": "/fields/System.Title", "value": title},
        ]
        if body:
            patch_ops.append({"op": "add", "path": "/fields/System.Description", "value": body})
        if assignee:
            patch_ops.append({"op": "add", "path": "/fields/System.AssignedTo", "value": assignee})
        if label:
            patch_ops.append({"op": "add", "path": "/fields/System.Tags", "value": label})

        resp = self._client.post(
            f"{self._wit_path()}/workitems/${quote(work_item_type, safe='')}",
            data=_json.dumps(patch_ops),
            headers={"Content-Type": "application/json-patch+json"},
        )
        return self._to_issue(resp.json())

    def get_issue(self, number: int) -> Issue:
        resp = self._client.get(f"{self._wit_path()}/workitems/{number}")
        return self._to_issue(resp.json())

    def close_issue(self, number: int) -> None:
        patch_ops = [{"op": "replace", "path": "/fields/System.State", "value": "Closed"}]
        self._client.patch(
            f"{self._wit_path()}/workitems/{number}",
            data=_json.dumps(patch_ops),
            headers={"Content-Type": "application/json-patch+json"},
        )

    def reopen_issue(self, number: int) -> None:
        patch_ops = [{"op": "replace", "path": "/fields/System.State", "value": "New"}]
        self._client.patch(
            f"{self._wit_path()}/workitems/{number}",
            data=_json.dumps(patch_ops),
            headers={"Content-Type": "application/json-patch+json"},
        )

    def delete_issue(self, number: int) -> None:
        self._client.delete(f"{self._wit_path()}/workitems/{number}")

    def list_issue_templates(self) -> list[IssueTemplate]:
        try:
            resp = self._client.get("/wit/workitemtypes")
        except (GfoError, requests.RequestException, KeyError, ValueError):
            return []
        templates: list[IssueTemplate] = []
        for t in resp.json().get("value", []):
            templates.append(
                IssueTemplate(
                    name=t.get("name") or "",
                    title="",
                    body=t.get("description") or "",
                    about=t.get("description") or "",
                    labels=tuple(),
                )
            )
        return templates

    # --- Repository ---

    def list_repositories(self, *, owner: str | None = None, limit: int = 30) -> list[Repository]:
        if owner is not None:
            raise NotSupportedError(
                self.service_name,
                "filtering repositories by owner "
                "(repositories are scoped to the configured project)",
            )
        results = paginate_top_skip(
            self._client,
            "/git/repositories",
            limit=limit,
            result_key="value",
        )
        return [self._to_repository(r, self._project) for r in results]

    def create_repository(
        self, *, name: str, private: bool = False, description: str = ""
    ) -> Repository:
        # project はベース URL に含まれるため payload には含めない
        payload = {"name": name}
        resp = self._client.post("/git/repositories", json=payload)
        return self._to_repository(resp.json(), self._project)

    def get_repository(self, owner: str | None = None, name: str | None = None) -> Repository:
        n = name if name is not None else self._repo
        resp = self._client.get(f"/git/repositories/{quote(n, safe='')}")
        return self._to_repository(resp.json(), self._project)

    def delete_repository(self) -> None:
        resp = self._client.get(self._git_path())
        try:
            repo_id = resp.json()["id"]
        except (KeyError, TypeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e
        self._client.delete(f"/git/repositories/{repo_id}")

    def update_repository(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        private: bool | None = None,
        default_branch: str | None = None,
    ) -> Repository:
        payload = {}
        if name is not None:
            payload["name"] = name
        if default_branch is not None:
            payload["defaultBranch"] = _add_refs_prefix(default_branch)
        resp = self._client.patch(self._git_path(), json=payload)
        return self._to_repository(resp.json(), self._project)

    def archive_repository(self) -> None:
        self._client.patch(self._git_path(), json={"isDisabled": True})

    def compare(self, base: str, head: str) -> CompareResult:
        resp = self._client.get(
            f"/git/repositories/{quote(self._repo, safe='')}/diffs/commits",
            params={"baseVersion": base, "targetVersion": head},
        )
        data = resp.json()
        changes = data.get("changes") or []
        files = tuple(
            CompareFile(
                filename=c.get("item", {}).get("path", "").lstrip("/"),
                status={
                    "add": "added",
                    "edit": "modified",
                    "delete": "deleted",
                    "rename": "renamed",
                }.get(c.get("changeType", "edit"), "modified"),
                additions=0,
                deletions=0,
            )
            for c in changes
        )
        return CompareResult(
            total_commits=0,
            ahead_by=data.get("aheadCount", 0),
            behind_by=data.get("behindCount", 0),
            files=files,
        )

    def migrate_repository(
        self,
        clone_url: str,
        name: str,
        *,
        private: bool = False,
        description: str = "",
        mirror: bool = False,
        auth_token: str | None = None,
    ) -> Repository:
        # まずリポジトリを作成
        repo = self.create_repository(name=name, private=private, description=description)
        # Import Request
        payload: dict = {"parameters": {"gitSource": {"url": clone_url}}}
        if auth_token:
            payload["parameters"]["gitSource"]["username"] = ""
            payload["parameters"]["gitSource"]["password"] = auth_token
        self._client.post(
            f"/git/repositories/{quote(name, safe='')}/importRequests",
            json=payload,
        )
        return repo

    # --- NotSupported ---

    def list_releases(self, *, limit: int = 30) -> list[Release]:
        raise NotSupportedError(self.service_name, "releases")

    def create_release(
        self,
        *,
        tag: str,
        title: str = "",
        notes: str = "",
        draft: bool = False,
        prerelease: bool = False,
        target: str | None = None,
    ) -> Release:
        raise NotSupportedError(self.service_name, "releases")

    def list_labels(self, *, limit: int = 0) -> list[Label]:
        raise NotSupportedError(self.service_name, "labels")

    def create_label(
        self, *, name: str, color: str | None = None, description: str | None = None
    ) -> Label:
        raise NotSupportedError(self.service_name, "labels")

    def list_milestones(self, *, limit: int = 0) -> list[Milestone]:
        raise NotSupportedError(self.service_name, "milestones")

    def create_milestone(
        self, *, title: str, description: str | None = None, due_date: str | None = None
    ) -> Milestone:
        raise NotSupportedError(self.service_name, "milestones")

    # --- 変換ヘルパー（追加型） ---

    @staticmethod
    def _to_comment(data: dict) -> Comment:
        from gfo.exceptions import GfoError

        try:
            # PR Threads の場合は comments 配列の最初のコメント
            if "comments" in data:
                comment_data = (data.get("comments") or [{}])[0]
            else:
                comment_data = data
            # PR thread は author, content フィールド / Work Item comment は createdBy, text フィールド
            author = comment_data.get("author") or comment_data.get("createdBy") or {}
            return Comment(
                id=comment_data.get("id") or data.get("id") or 0,
                body=comment_data.get("content") or comment_data.get("text") or "",
                author=author.get("uniqueName") or author.get("displayName") or "",
                url="",
                created_at=comment_data.get("publishedDate") or "",
                updated_at=comment_data.get("lastUpdatedDate"),
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    @staticmethod
    def _to_review(data: dict) -> Review:
        from gfo.exceptions import GfoError

        try:
            vote_map = {
                10: "approved",
                5: "approved",
                0: "commented",
                -5: "changes_requested",
                -10: "changes_requested",
            }
            vote = data.get("vote", 0)
            return Review(
                id=data.get("id") or 0,
                state=vote_map.get(vote, "commented"),
                body="",
                author=data.get("uniqueName") or data.get("displayName") or "",
                url="",
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    @staticmethod
    def _to_branch(data: dict) -> Branch:
        from gfo.exceptions import GfoError

        try:
            # AzDO: name は "refs/heads/xxx" 形式
            name = data["name"]
            if name.startswith("refs/heads/"):
                name = name[len("refs/heads/") :]
            return Branch(
                name=name,
                sha=(data.get("objectId") or ""),
                protected=False,
                url="",
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    @staticmethod
    def _to_tag(data: dict) -> Tag:
        from gfo.exceptions import GfoError

        try:
            name = data.get("name", "")
            if name.startswith("refs/tags/"):
                name = name[len("refs/tags/") :]
            return Tag(
                name=name,
                sha=data.get("objectId") or "",
                message="",
                url="",
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    @staticmethod
    def _to_commit_status(data: dict) -> CommitStatus:
        from gfo.exceptions import GfoError

        try:
            state_map = {
                "succeeded": "success",
                "failed": "failure",
                "pending": "pending",
                "error": "error",
                "notApplicable": "pending",
                "notSet": "pending",
            }
            state_obj = data.get("state", {})
            raw = state_obj if isinstance(state_obj, str) else state_obj.get("state", "pending")
            genre = (data.get("context") or {}).get("genre") or ""
            name = (data.get("context") or {}).get("name") or ""
            context = f"{genre}/{name}" if genre else name
            return CommitStatus(
                state=state_map.get(raw, raw),
                context=context,
                description=data.get("description") or "",
                target_url=data.get("targetUrl") or "",
                created_at=data.get("creationDate") or "",
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    @staticmethod
    def _to_webhook(data: dict) -> Webhook:
        from gfo.exceptions import GfoError

        try:
            events = tuple([data.get("eventType") or ""] if data.get("eventType") else [])
            return Webhook(
                id=data.get("id") or 0,
                url=(data.get("consumerInputs") or {}).get("url") or "",
                events=events,
                active=data.get("status") == "enabled",
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    @staticmethod
    def _to_pipeline(data: dict) -> Pipeline:
        from gfo.exceptions import GfoError

        try:
            status_map = {
                "completed": lambda d: {
                    "succeeded": "success",
                    "failed": "failure",
                    "canceled": "cancelled",
                }.get(d.get("result", ""), "failure"),
                "inProgress": lambda d: "running",
                "notStarted": lambda d: "pending",
                "cancelling": lambda d: "cancelled",
            }
            raw_status = data.get("status", "notStarted")
            mapper = status_map.get(raw_status)
            status = mapper(data) if mapper else raw_status
            source_branch = _strip_refs_prefix(data.get("sourceBranch") or "")
            return Pipeline(
                id=data["id"],
                status=status,
                ref=source_branch,
                url=data.get("_links", {}).get("web", {}).get("href") or "",
                created_at=data.get("queueTime") or "",
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    # --- Comment ---

    def list_comments(self, resource: str, number: int, *, limit: int = 30) -> list[Comment]:
        if resource == "pr":
            resp = self._client.get(f"{self._git_path()}/pullrequests/{number}/threads")
            threads = resp.json().get("value", [])
            result = []
            for thread in threads:
                for comment in thread.get("comments") or []:
                    result.append(self._to_comment(comment))
                    if limit > 0 and len(result) >= limit:
                        return result
            return result
        else:
            # Work Item コメントは 7.1-preview.3 が必要
            resp = self._client.get(
                f"{self._wit_path()}/workitems/{number}/comments",
                params={"api-version": "7.1-preview.3"},
            )
            comments_data = resp.json().get("comments", [])
            return [self._to_comment(c) for c in comments_data[: limit if limit > 0 else None]]

    def create_comment(self, resource: str, number: int, *, body: str) -> Comment:
        if resource == "pr":
            # PR コメントは Thread を作成する
            payload = {
                "comments": [{"parentCommentId": 0, "content": body, "commentType": 1}],
                "status": 1,
            }
            resp = self._client.post(
                f"{self._git_path()}/pullrequests/{number}/threads", json=payload
            )
            return self._to_comment(resp.json())
        else:
            payload = {"text": body}
            # Work Item コメントは 7.1-preview.3 が必要
            resp = self._client.post(
                f"{self._wit_path()}/workitems/{number}/comments",
                json=payload,
                params={"api-version": "7.1-preview.3"},
            )
            return self._to_comment(resp.json())

    def update_comment(self, resource: str, comment_id: int, *, body: str) -> Comment:
        # Azure DevOps のコメント更新は thread_id と comment_id の両方が必要で、comment_id のみでは不可
        raise NotSupportedError(
            self.service_name, "comment update (requires thread ID for PR comments)"
        )

    def delete_comment(self, resource: str, comment_id: int) -> None:
        raise NotSupportedError(
            self.service_name, "comment delete (requires thread ID for PR comments)"
        )

    # --- PR update ---

    def update_pull_request(
        self,
        number: int,
        *,
        title: str | None = None,
        body: str | None = None,
        base: str | None = None,
        add_labels: list[str] | None = None,
        remove_labels: list[str] | None = None,
        add_assignees: list[str] | None = None,
        remove_assignees: list[str] | None = None,
        milestone: str | None = None,
    ) -> PullRequest:
        self._warn_unsupported_params(
            "pr edit",
            add_labels=add_labels,
            remove_labels=remove_labels,
            add_assignees=add_assignees,
            remove_assignees=remove_assignees,
            milestone=milestone,
        )
        payload: dict = {}
        if title is not None:
            payload["title"] = title
        if body is not None:
            payload["description"] = body
        if base is not None:
            payload["targetRefName"] = _add_refs_prefix(base)
        resp = self._client.patch(f"{self._git_path()}/pullrequests/{number}", json=payload)
        return self._to_pull_request(resp.json())

    # --- Issue update (Work Item patch) ---

    def update_issue(
        self,
        number: int,
        *,
        title: str | None = None,
        body: str | None = None,
        assignee: str | None = None,
        label: str | None = None,
        add_labels: list[str] | None = None,
        remove_labels: list[str] | None = None,
        add_assignees: list[str] | None = None,
        remove_assignees: list[str] | None = None,
        milestone: str | None = None,
    ) -> Issue:
        self._warn_unsupported_params(
            "issue edit",
            add_labels=add_labels,
            remove_labels=remove_labels,
            add_assignees=add_assignees,
            remove_assignees=remove_assignees,
            milestone=milestone,
        )
        patch_ops = []
        if title is not None:
            patch_ops.append({"op": "replace", "path": "/fields/System.Title", "value": title})
        if body is not None:
            patch_ops.append({"op": "replace", "path": "/fields/System.Description", "value": body})
        if assignee is not None:
            patch_ops.append(
                {"op": "replace", "path": "/fields/System.AssignedTo", "value": assignee}
            )
        if label is not None:
            patch_ops.append({"op": "replace", "path": "/fields/System.Tags", "value": label})
        resp = self._client.patch(
            f"{self._wit_path()}/workitems/{number}",
            data=_json.dumps(patch_ops),
            headers={"Content-Type": "application/json-patch+json"},
        )
        return self._to_issue(resp.json())

    # --- Review (Reviewers API) ---

    def list_reviews(self, number: int) -> list[Review]:
        resp = self._client.get(f"{self._git_path()}/pullrequests/{number}/reviewers")
        reviewers = resp.json()
        if not isinstance(reviewers, list):
            reviewers = reviewers.get("value", [])
        return [self._to_review(r) for r in reviewers]

    def _connection_data_url(self) -> str:
        """組織レベルの connectionData エンドポイント URL を返す。

        base_url は `https://dev.azure.com/{org}/{project}/_apis` のようなプロジェクト
        スコープ URL だが、connectionData は組織スコープ
        (`https://dev.azure.com/{org}/_apis/connectionData`) で呼ぶ必要がある。
        """
        parsed = urlparse(self._client.base_url)
        return f"{parsed.scheme}://{parsed.netloc}/{self._org}/_apis/connectionData"

    def create_review(self, number: int, *, state: str, body: str = "") -> Review:
        vote_map = {"APPROVE": 10, "REQUEST_CHANGES": -10, "COMMENT": 0}
        vote = vote_map.get(state.upper(), 0)
        # connectionData は組織スコープのため get_absolute で組織レベル URL を直接指定
        # connectionData は api-version=7.1-preview が必要（7.1 だと 400 エラー）
        user_resp = self._client.get_absolute(
            self._connection_data_url(), params={"api-version": "7.1-preview"}
        )
        user_id = (user_resp.json().get("authenticatedUser") or {}).get("id") or ""
        payload = {"vote": vote}
        resp = self._client.put(
            f"{self._git_path()}/pullrequests/{number}/reviewers/{user_id}",
            json=payload,
        )
        return self._to_review(resp.json())

    # --- Branch ---

    def get_branch(self, name: str) -> Branch:
        resp = self._client.get(
            f"{self._git_path()}/refs",
            params={"filter": f"heads/{name}"},
        )
        items = resp.json().get("value", [])
        items = [i for i in items if i.get("name") == f"refs/heads/{name}"]
        if not items:
            from gfo.exceptions import NotFoundError

            raise NotFoundError(f"refs/heads/{name}")
        return self._to_branch(items[0])

    def list_branches(self, *, limit: int = 30) -> list[Branch]:
        results = paginate_top_skip(
            self._client,
            f"{self._git_path()}/refs",
            params={"filter": "heads/"},
            limit=limit,
            result_key="value",
        )
        return [self._to_branch(r) for r in results]

    def create_branch(self, *, name: str, ref: str) -> Branch:
        # AzDO refs 更新 API でブランチを作成（pushes API は commit が必要なので使用しない）
        payload = [{"name": f"refs/heads/{name}", "oldObjectId": "0" * 40, "newObjectId": ref}]
        self._client.post(f"{self._git_path()}/refs", json=payload)
        resp = self._client.get(
            f"{self._git_path()}/refs",
            params={"filter": f"heads/{name}"},
        )
        items = resp.json().get("value", [])
        if not items:
            from gfo.exceptions import GfoError

            raise GfoError(f"Branch '{name}' not found after creation")
        return self._to_branch(items[0])

    def delete_branch(self, *, name: str) -> None:
        # 現在の SHA を取得
        resp = self._client.get(
            f"{self._git_path()}/refs",
            params={"filter": f"heads/{name}"},
        )
        items = resp.json().get("value", [])
        if not items:
            from gfo.exceptions import NotFoundError

            raise NotFoundError(f"refs/heads/{name}")
        sha = items[0]["objectId"]
        payload = [{"name": f"refs/heads/{name}", "oldObjectId": sha, "newObjectId": "0" * 40}]
        self._client.post(f"{self._git_path()}/refs", json=payload)

    # --- Tag ---

    def get_tag(self, name: str) -> Tag:
        resp = self._client.get(
            f"{self._git_path()}/refs",
            params={"filter": f"tags/{name}"},
        )
        items = resp.json().get("value", [])
        items = [i for i in items if i.get("name") == f"refs/tags/{name}"]
        if not items:
            from gfo.exceptions import NotFoundError

            raise NotFoundError(f"refs/tags/{name}")
        return self._to_tag(items[0])

    def list_tags(self, *, limit: int = 30) -> list[Tag]:
        results = paginate_top_skip(
            self._client,
            f"{self._git_path()}/refs",
            params={"filter": "tags/"},
            limit=limit,
            result_key="value",
        )
        return [self._to_tag(r) for r in results]

    def create_tag(self, *, name: str, ref: str, message: str = "") -> Tag:
        # Lightweight tag: refs 更新 API でタグを作成
        payload = [{"name": f"refs/tags/{name}", "oldObjectId": "0" * 40, "newObjectId": ref}]
        self._client.post(f"{self._git_path()}/refs", json=payload)
        return Tag(name=name, sha=ref, message=message, url="")

    def delete_tag(self, *, name: str) -> None:
        resp = self._client.get(
            f"{self._git_path()}/refs",
            params={"filter": f"tags/{name}"},
        )
        items = resp.json().get("value", [])
        if not items:
            from gfo.exceptions import NotFoundError

            raise NotFoundError(f"refs/tags/{name}")
        sha = items[0]["objectId"]
        payload = [{"name": f"refs/tags/{name}", "oldObjectId": sha, "newObjectId": "0" * 40}]
        self._client.post(f"{self._git_path()}/refs", json=payload)

    # --- CommitStatus ---

    def list_commit_statuses(self, ref: str, *, limit: int = 30) -> list[CommitStatus]:
        resp = self._client.get(f"{self._git_path()}/commits/{quote(ref, safe='')}/statuses")
        statuses = resp.json()
        if isinstance(statuses, dict):
            statuses = statuses.get("value", [])
        return [self._to_commit_status(s) for s in statuses[: limit if limit > 0 else None]]

    def create_commit_status(
        self,
        ref: str,
        *,
        state: str,
        context: str = "",
        description: str = "",
        target_url: str = "",
    ) -> CommitStatus:
        state_map = {
            "success": "succeeded",
            "failure": "failed",
            "pending": "pending",
            "error": "error",
        }
        parts = context.split("/", 1) if "/" in context else ["gfo", context]
        payload: dict = {
            "state": state_map.get(state, state),
            "context": {"name": parts[-1], "genre": parts[0] if len(parts) > 1 else "gfo"},
        }
        if description:
            payload["description"] = description
        if target_url:
            payload["targetUrl"] = target_url
        resp = self._client.post(
            f"{self._git_path()}/commits/{quote(ref, safe='')}/statuses",
            json=payload,
        )
        return self._to_commit_status(resp.json())

    # --- File (Azure DevOps Items API) ---

    def get_file_content(self, path: str, *, ref: str | None = None) -> tuple[str, str]:
        params: dict = {"path": path, "includeContent": "true", "$format": "json"}
        if ref is not None:
            params["versionDescriptor.version"] = ref
            # 40文字の16進数は commit SHA、それ以外はブランチ名として扱う
            is_sha = len(ref) == 40 and all(c in "0123456789abcdef" for c in ref)
            params["versionDescriptor.versionType"] = "commit" if is_sha else "branch"
        resp = self._client.get(f"{self._git_path()}/items", params=params)
        data = resp.json()
        try:
            content = data.get("content") or ""
            # objectId は blob SHA
            sha = data.get("objectId") or ""
        except (KeyError, TypeError, AttributeError) as e:
            from gfo.exceptions import GfoError

            raise GfoError(f"Unexpected API response: {e}") from e
        return content, sha

    def create_or_update_file(
        self,
        path: str,
        *,
        content: str,
        message: str,
        sha: str | None = None,
        branch: str | None = None,
    ) -> str | None:
        # AzDO ではファイル作成/更新に pushes API を使う
        branch = branch or "main"
        # 既存ブランチの最新コミット SHA を取得
        refs_resp = self._client.get(
            f"{self._git_path()}/refs",
            params={"filter": f"heads/{branch}"},
        )
        ref_items = refs_resp.json().get("value", [])
        old_oid = ref_items[0]["objectId"] if ref_items else "0" * 40
        change_type = 2 if sha else 1  # 1=add, 2=edit
        payload = {
            "refUpdates": [{"name": f"refs/heads/{branch}", "oldObjectId": old_oid}],
            "commits": [
                {
                    "comment": message,
                    "changes": [
                        {
                            "changeType": change_type,
                            "item": {"path": f"/{path}" if not path.startswith("/") else path},
                            "newContent": {
                                "content": base64.b64encode(content.encode()).decode(),
                                "contentType": "base64encoded",
                            },
                        }
                    ],
                }
            ],
        }
        resp = self._client.post(f"{self._git_path()}/pushes", json=payload)
        commits = resp.json().get("commits", [])
        return commits[0].get("commitId") if commits else None

    def delete_file(
        self,
        path: str,
        *,
        sha: str,
        message: str,
        branch: str | None = None,
    ) -> None:
        branch = branch or "main"
        refs_resp = self._client.get(
            f"{self._git_path()}/refs",
            params={"filter": f"heads/{branch}"},
        )
        ref_items = refs_resp.json().get("value", [])
        old_oid = ref_items[0]["objectId"] if ref_items else "0" * 40
        payload = {
            "refUpdates": [{"name": f"refs/heads/{branch}", "oldObjectId": old_oid}],
            "commits": [
                {
                    "comment": message,
                    "changes": [
                        {
                            "changeType": 16,  # delete
                            "item": {"path": f"/{path}" if not path.startswith("/") else path},
                        }
                    ],
                }
            ],
        }
        self._client.post(f"{self._git_path()}/pushes", json=payload)

    # --- Webhook (Service Hooks) ---

    def list_webhooks(self, *, limit: int = 30) -> list[Webhook]:
        raise NotSupportedError(self.service_name, "webhook list (use Azure DevOps Service Hooks)")

    def create_webhook(self, *, url: str, events: list[str], secret: str | None = None) -> Webhook:
        raise NotSupportedError(
            self.service_name, "webhook create (use Azure DevOps Service Hooks)"
        )

    def delete_webhook(self, *, hook_id: int) -> None:
        raise NotSupportedError(
            self.service_name, "webhook delete (use Azure DevOps Service Hooks)"
        )

    # --- DeployKey ---

    def list_deploy_keys(self, *, limit: int = 30) -> list[DeployKey]:
        raise NotSupportedError(
            self.service_name, "deploy-key (Azure DevOps uses SSH keys at org level)"
        )

    def create_deploy_key(self, *, title: str, key: str, read_only: bool = True) -> DeployKey:
        raise NotSupportedError(
            self.service_name, "deploy-key (Azure DevOps uses SSH keys at org level)"
        )

    def delete_deploy_key(self, *, key_id: int) -> None:
        raise NotSupportedError(
            self.service_name, "deploy-key (Azure DevOps uses SSH keys at org level)"
        )

    # --- Collaborator ---

    def list_collaborators(self, *, limit: int = 30) -> list[str]:
        raise NotSupportedError(
            self.service_name, "collaborator list (Azure DevOps uses teams/members)"
        )

    def add_collaborator(self, *, username: str, permission: str = "write") -> None:
        raise NotSupportedError(
            self.service_name, "collaborator add (Azure DevOps uses teams/members)"
        )

    def remove_collaborator(self, *, username: str) -> None:
        raise NotSupportedError(
            self.service_name, "collaborator remove (Azure DevOps uses teams/members)"
        )

    # --- Pipeline (Build API) ---

    def list_pipelines(self, *, ref: str | None = None, limit: int = 30) -> list[Pipeline]:
        params: dict = {}
        if ref is not None:
            params["branchName"] = f"refs/heads/{ref}" if not ref.startswith("refs/") else ref
        results = paginate_top_skip(
            self._client,
            "/build/builds",
            params=params,
            limit=limit,
            result_key="value",
        )
        return [self._to_pipeline(r) for r in results]

    def get_pipeline(self, pipeline_id: int | str) -> Pipeline:
        resp = self._client.get(f"/build/builds/{pipeline_id}")
        return self._to_pipeline(resp.json())

    def cancel_pipeline(self, pipeline_id: int | str) -> None:
        self._client.patch(
            f"/build/builds/{pipeline_id}",
            data=_json.dumps([{"op": "replace", "path": "/status", "value": "cancelling"}]),
            headers={"Content-Type": "application/json-patch+json"},
        )

    def trigger_pipeline(
        self, ref: str, *, workflow: str | None = None, inputs: dict | None = None
    ) -> Pipeline:
        payload: dict = {
            "sourceBranch": ref if ref.startswith("refs/") else f"refs/heads/{ref}",
        }
        if workflow:
            payload["definition"] = {"id": int(workflow)}
        resp = self._client.post("/build/builds", json=payload)
        return self._to_pipeline(resp.json())

    def retry_pipeline(self, pipeline_id: int | str) -> Pipeline:
        resp = self._client.patch(
            f"/build/builds/{pipeline_id}",
            json={"retry": True},
        )
        return self._to_pipeline(resp.json())

    def get_pipeline_logs(self, pipeline_id: int | str, *, job_id: int | str | None = None) -> str:
        if job_id is not None:
            resp = self._client.get(f"/build/builds/{pipeline_id}/logs/{job_id}")
            return str(resp.text)
        logs_resp = self._client.get(f"/build/builds/{pipeline_id}/logs")
        log_entries = logs_resp.json().get("value", [])
        logs = []
        for entry in log_entries:
            log_id = entry.get("id", "")
            try:
                resp = self._client.get(f"/build/builds/{pipeline_id}/logs/{log_id}")
                logs.append(f"=== Log {log_id} ===\n{resp.text}")
            except (GfoError, requests.RequestException):
                logs.append(f"=== Log {log_id} ===\n(log unavailable)")
        return "\n".join(logs)

    # --- User ---

    def get_current_user(self) -> dict:
        # /_apis/profile/profiles/me は組織スコープ外のため connectionData を使用
        # connectionData は組織スコープのため get_absolute で組織レベル URL を直接指定
        # connectionData は api-version=7.1-preview が必要（7.1 だと 400 エラー）
        resp = self._client.get_absolute(
            self._connection_data_url(), params={"api-version": "7.1-preview"}
        )
        data = resp.json()
        user = data.get("authenticatedUser") or {}
        return {
            "id": user.get("id", ""),
            "displayName": user.get("providerDisplayName", ""),
            "login": user.get("providerDisplayName", ""),
        }

    # --- BranchProtection (Policy) ---

    def list_branch_protections(self, *, limit: int = 30) -> list[BranchProtection]:
        raise NotSupportedError(
            self.service_name,
            "branch-protect list (Azure DevOps uses policy configurations)",
        )

    def get_branch_protection(self, branch: str) -> BranchProtection:
        raise NotSupportedError(
            self.service_name,
            "branch-protect view (Azure DevOps uses policy configurations)",
        )

    def set_branch_protection(
        self,
        branch: str,
        *,
        require_reviews: int | None = None,
        require_status_checks: list[str] | None = None,
        enforce_admins: bool | None = None,
        allow_force_push: bool | None = None,
        allow_deletions: bool | None = None,
    ) -> BranchProtection:
        raise NotSupportedError(
            self.service_name,
            "branch-protect set (Azure DevOps uses policy configurations)",
        )

    def remove_branch_protection(self, branch: str) -> None:
        raise NotSupportedError(
            self.service_name,
            "branch-protect remove (Azure DevOps uses policy configurations)",
        )

    # --- Organization (Project) ---

    def _org_api_url(self, path: str) -> str:
        """組織レベル API の絶対 URL を構築する。

        Projects API 等はプロジェクト配下ではなく組織直下のエンドポイントを使用する。
        base_url: https://dev.azure.com/{org}/{project}/_apis
        必要:     https://dev.azure.com/{org}/_apis{path}
        """
        base: str = self._client.base_url
        # /{project}/_apis を /_apis に置換
        project_segment = f"/{quote(self._project, safe='')}"
        idx = base.find(project_segment + "/_apis")
        if idx >= 0:
            return base[:idx] + "/_apis" + path
        return base + path

    def list_organizations(self, *, limit: int = 30) -> list[Organization]:
        url = self._org_api_url("/projects")
        resp = self._client.get_absolute(url, params={"$top": str(limit), "api-version": "7.1"})
        data = resp.json()
        results = data.get("value", [])
        return [self._to_organization(d) for d in results]

    def get_organization(self, name: str) -> Organization:
        url = self._org_api_url(f"/projects/{quote(name, safe='')}")
        resp = self._client.get_absolute(url, params={"api-version": "7.1"})
        return self._to_organization(resp.json())

    def list_org_members(self, name: str, *, limit: int = 30) -> list[str]:
        raise NotSupportedError(
            self.service_name,
            "org members (Azure DevOps uses teams for member management)",
        )

    def list_org_repos(self, name: str, *, limit: int = 30) -> list[Repository]:
        results = paginate_top_skip(
            self._client,
            "/git/repositories",
            limit=0,
            result_key="value",
        )
        filtered = [r for r in results if (r.get("project") or {}).get("name") == name]
        return [self._to_repository(r, name) for r in filtered[: limit if limit > 0 else None]]

    def _to_organization(self, data: dict) -> Organization:
        try:
            name = data.get("name") or ""
            # _links.web.href を優先し、なければ dev.azure.com URL を構築
            web_url = ""
            links = data.get("_links") or {}
            web = links.get("web") or {}
            if web.get("href"):
                web_url = web["href"]
            elif name:
                web_url = f"https://dev.azure.com/{self._org}/{name}"
            return Organization(
                name=name,
                display_name=name,
                description=data.get("description"),
                url=web_url,
            )
        except (KeyError, TypeError, AttributeError) as e:
            raise GfoError(f"Unexpected API response: missing field {e}") from e

    # --- Browse ---

    def get_web_url(self, resource: str = "repo", number: int | str | None = None) -> str:
        base = f"https://dev.azure.com/{self._org}/{self._project}/_git/{self._repo}"
        if resource == "pr":
            return f"{base}/pullrequests" if number is None else f"{base}/pullrequest/{number}"
        if resource == "issue":
            proj_base = f"https://dev.azure.com/{self._org}/{self._project}"
            return (
                f"{proj_base}/_workitems"
                if number is None
                else f"{proj_base}/_workitems?id={number}"
            )
        if resource == "release":
            raise NotSupportedError(self.service_name, "browse release")
        if resource == "milestone":
            raise NotSupportedError(self.service_name, "browse milestone")
        if resource == "settings":
            return f"https://dev.azure.com/{self._org}/{self._project}/_settings/repositories"
        return base

    # --- Search ---

    def search_repositories(self, query: str, *, limit: int = 30) -> list[Repository]:
        # AzDO でリポジトリ検索は /git/repositories でフィルタ
        results = paginate_top_skip(
            self._client,
            "/git/repositories",
            limit=0,
            result_key="value",
        )
        filtered = [r for r in results if query.lower() in r.get("name", "").lower()]
        return [
            self._to_repository(r, self._project) for r in filtered[: limit if limit > 0 else None]
        ]

    def search_issues(self, query: str, *, limit: int = 30) -> list[Issue]:
        # WIQL を使って検索
        wiql = f"SELECT [System.Id] FROM WorkItems WHERE [System.TeamProject] = '{_wiql_escape(self._project)}' AND [System.Title] CONTAINS '{_wiql_escape(query)}'"  # nosec B608
        wiql_params = {"$top": limit} if limit > 0 else {"$top": 200}
        wiql_resp = self._client.post(
            f"{self._wit_path()}/wiql",
            json={"query": wiql},
            params=wiql_params,
        )
        wiql_body = wiql_resp.json()
        if not isinstance(wiql_body, dict):
            from gfo.exceptions import GfoError

            raise GfoError(f"Unexpected API response: {type(wiql_body)}")
        try:
            ids = [wi["id"] for wi in wiql_body.get("workItems", [])]
        except (KeyError, TypeError) as e:
            from gfo.exceptions import GfoError

            raise GfoError(f"Unexpected API response: {e}") from e
        if not ids:
            return []
        ids_str = ",".join(str(x) for x in ids[: limit if limit > 0 else None])
        resp = self._client.get(
            f"{self._wit_path()}/workitems",
            params={"ids": ids_str, "$expand": "None"},
        )
        batch_body = resp.json()
        if not isinstance(batch_body, dict):
            from gfo.exceptions import GfoError

            raise GfoError(f"Unexpected API response: {type(batch_body)}")
        return [self._to_issue(item) for item in batch_body.get("value", [])]

    # --- Issue Dependencies (Work Item Links) ---

    def list_issue_dependencies(self, number):
        resp = self._client.get(
            f"{self._wit_path()}/workitems/{number}", params={"$expand": "relations"}
        )
        data = resp.json()
        deps = []
        for rel in data.get("relations") or []:
            if "System.LinkTypes.Dependency" in (rel.get("rel") or ""):
                url = rel.get("url") or ""
                dep_id = int(url.rsplit("/", 1)[-1]) if "/" in url else 0
                if dep_id:
                    dep = self.get_issue(dep_id)
                    deps.append(dep)
        return deps

    def add_issue_dependency(self, number, depends_on):
        parsed = urlparse(self._client.base_url)
        base = f"{parsed.scheme}://{parsed.hostname}"
        if parsed.port:
            base += f":{parsed.port}"
        target_url = f"{base}/{self._org}/{self._project}/_apis/wit/workitems/{depends_on}"
        patch = [
            {
                "op": "add",
                "path": "/relations/-",
                "value": {"rel": "System.LinkTypes.Dependency-Forward", "url": target_url},
            }
        ]
        self._client.patch(
            f"{self._wit_path()}/workitems/{number}",
            json=patch,
            headers={"Content-Type": "application/json-patch+json"},
        )

    def remove_issue_dependency(self, number, depends_on):
        resp = self._client.get(
            f"{self._wit_path()}/workitems/{number}", params={"$expand": "relations"}
        )
        data = resp.json()
        for i, rel in enumerate(data.get("relations") or []):
            if "System.LinkTypes.Dependency" in (rel.get("rel") or ""):
                url = rel.get("url") or ""
                dep_id = int(url.rsplit("/", 1)[-1]) if "/" in url else 0
                if dep_id == depends_on:
                    patch = [{"op": "remove", "path": f"/relations/{i}"}]
                    self._client.patch(
                        f"{self._wit_path()}/workitems/{number}",
                        json=patch,
                        headers={"Content-Type": "application/json-patch+json"},
                    )
                    return

    # --- Issue Timeline (Work Item Updates) ---

    def get_issue_timeline(self, number, *, limit=30):
        resp = self._client.get(f"{self._wit_path()}/workitems/{number}/updates")
        updates = resp.json().get("value") or []
        events = []
        for u in updates[:limit]:
            fields = u.get("fields") or {}
            detail_parts = []
            for field_name, change in fields.items():
                new_val = change.get("newValue") or ""
                if new_val:
                    detail_parts.append(f"{field_name}: {new_val}")
            if detail_parts:
                events.append(
                    TimelineEvent(
                        id=u.get("id") or 0,
                        event="updated",
                        actor=(u.get("revisedBy") or {}).get("displayName") or "",
                        created_at=(
                            u.get("revisedDate")
                            or u.get("fields", {}).get("System.ChangedDate", {}).get("newValue")
                            or ""
                        ),
                        detail="; ".join(detail_parts[:3]),
                    )
                )
        return events

    # --- Search PRs ---

    def search_pull_requests(self, query, *, state=None, limit=30):
        params = {"searchCriteria.includeLinks": "false"}
        if state and state != "all":
            api_state = _PR_STATE_TO_API.get(state, state)
            params["searchCriteria.status"] = api_state
        results = paginate_top_skip(
            self._client, f"{self._git_path()}/pullrequests", params=params, limit=limit
        )
        prs = [self._to_pull_request(r) for r in results]
        if query:
            query_lower = query.lower()
            prs = [
                p
                for p in prs
                if query_lower in p.title.lower() or (p.body and query_lower in p.body.lower())
            ]
        return prs

    # --- Search Commits ---

    def search_commits(self, query, *, author=None, since=None, until=None, limit=30):
        params = {}
        if author:
            params["searchCriteria.author"] = author
        if since:
            params["searchCriteria.fromDate"] = since
        if until:
            params["searchCriteria.toDate"] = until
        results = paginate_top_skip(
            self._client, f"{self._git_path()}/commits", params=params, limit=limit
        )
        commits = [
            Commit(
                sha=c.get("commitId") or "",
                message=c.get("comment") or "",
                author=(c.get("author") or {}).get("name") or "",
                url=c.get("remoteUrl") or c.get("url") or "",
                created_at=(c.get("author") or {}).get("date") or "",
            )
            for c in results
        ]
        if query:
            query_lower = query.lower()
            commits = [c for c in commits if query_lower in c.message.lower()]
        return commits

    # --- Time Tracking (Completed Work) ---

    def add_time_entry(self, issue_number, duration):
        hours = duration / 3600
        patch = [
            {"op": "add", "path": "/fields/Microsoft.VSTS.Scheduling.CompletedWork", "value": hours}
        ]
        self._client.patch(
            f"{self._wit_path()}/workitems/{issue_number}",
            json=patch,
            headers={"Content-Type": "application/json-patch+json"},
        )
        return TimeEntry(id=0, user="", duration=duration, created_at="")



"""全アダプターの BranchProtection テスト。"""

from __future__ import annotations

import json

import pytest
import responses

from gfo.exceptions import AuthenticationError, NotFoundError, NotSupportedError

# --- GitHub ---


class TestGitHubBranchProtect:
    @responses.activate
    def test_get(self, github_adapter):
        responses.add(
            responses.GET,
            "https://api.github.com/repos/test-owner/test-repo/branches/main/protection",
            json={
                "required_pull_request_reviews": {"required_approving_review_count": 2},
                "required_status_checks": {"strict": True, "contexts": ["ci/test"]},
                "enforce_admins": {"enabled": True},
                "restrictions": None,
                "allow_force_pushes": {"enabled": False},
                "allow_deletions": {"enabled": False},
            },
        )
        bp = github_adapter.get_branch_protection("main")
        assert bp.branch == "main"
        assert bp.require_reviews == 2
        assert bp.require_status_checks == ("ci/test",)
        assert bp.enforce_admins is True
        assert bp.allow_force_push is False

    @responses.activate
    def test_set(self, github_adapter):
        # GET 現在値
        responses.add(
            responses.GET,
            "https://api.github.com/repos/test-owner/test-repo/branches/main/protection",
            json={
                "required_pull_request_reviews": None,
                "required_status_checks": None,
                "enforce_admins": {"enabled": False},
                "allow_force_pushes": {"enabled": False},
                "allow_deletions": {"enabled": False},
            },
        )
        # PUT
        responses.add(
            responses.PUT,
            "https://api.github.com/repos/test-owner/test-repo/branches/main/protection",
            json={
                "required_pull_request_reviews": {"required_approving_review_count": 1},
                "required_status_checks": None,
                "enforce_admins": {"enabled": False},
                "allow_force_pushes": {"enabled": False},
                "allow_deletions": {"enabled": False},
            },
        )
        bp = github_adapter.set_branch_protection("main", require_reviews=1)
        assert bp.require_reviews == 1
        req_body = json.loads(responses.calls[1].request.body)
        assert req_body["required_pull_request_reviews"]["required_approving_review_count"] == 1

    @responses.activate
    def test_remove(self, github_adapter):
        responses.add(
            responses.DELETE,
            "https://api.github.com/repos/test-owner/test-repo/branches/main/protection",
            status=204,
        )
        github_adapter.remove_branch_protection("main")

    @responses.activate
    def test_list(self, github_adapter):
        responses.add(
            responses.GET,
            "https://api.github.com/repos/test-owner/test-repo/branches",
            json=[
                {"name": "main", "protected": True},
                {"name": "dev", "protected": False},
            ],
        )
        responses.add(
            responses.GET,
            "https://api.github.com/repos/test-owner/test-repo/branches/main/protection",
            json={
                "required_pull_request_reviews": None,
                "required_status_checks": None,
                "enforce_admins": False,
                "allow_force_pushes": False,
                "allow_deletions": False,
            },
        )
        bps = github_adapter.list_branch_protections()
        assert len(bps) == 1
        assert bps[0].branch == "main"

    @responses.activate
    def test_list_empty(self, github_adapter):
        responses.add(
            responses.GET,
            "https://api.github.com/repos/test-owner/test-repo/branches",
            json=[],
        )
        assert github_adapter.list_branch_protections() == []

    @responses.activate
    def test_get_404(self, github_adapter):
        responses.add(
            responses.GET,
            "https://api.github.com/repos/test-owner/test-repo/branches/main/protection",
            status=404,
        )
        with pytest.raises(NotFoundError):
            github_adapter.get_branch_protection("main")

    @responses.activate
    def test_remove_403(self, github_adapter):
        responses.add(
            responses.DELETE,
            "https://api.github.com/repos/test-owner/test-repo/branches/main/protection",
            status=403,
        )
        with pytest.raises(AuthenticationError):
            github_adapter.remove_branch_protection("main")


# --- GitLab ---


class TestGitLabBranchProtect:
    @responses.activate
    def test_list(self, gitlab_adapter):
        responses.add(
            responses.GET,
            "https://gitlab.com/api/v4/projects/test-owner%2Ftest-repo/protected_branches",
            json=[{"name": "main", "allow_force_push": False, "required_approvals": 1}],
        )
        bps = gitlab_adapter.list_branch_protections()
        assert len(bps) == 1
        assert bps[0].branch == "main"
        assert bps[0].require_reviews == 1

    @responses.activate
    def test_get(self, gitlab_adapter):
        responses.add(
            responses.GET,
            "https://gitlab.com/api/v4/projects/test-owner%2Ftest-repo/protected_branches/main",
            json={"name": "main", "allow_force_push": True, "required_approvals": 0},
        )
        bp = gitlab_adapter.get_branch_protection("main")
        assert bp.allow_force_push is True

    @responses.activate
    def test_set(self, gitlab_adapter):
        responses.add(
            responses.POST,
            "https://gitlab.com/api/v4/projects/test-owner%2Ftest-repo/protected_branches",
            json={"name": "main", "allow_force_push": False, "required_approvals": 2},
        )
        bp = gitlab_adapter.set_branch_protection("main", require_reviews=2)
        assert bp.require_reviews == 2

    @responses.activate
    def test_remove(self, gitlab_adapter):
        responses.add(
            responses.DELETE,
            "https://gitlab.com/api/v4/projects/test-owner%2Ftest-repo/protected_branches/main",
            status=204,
        )
        gitlab_adapter.remove_branch_protection("main")

    @responses.activate
    def test_list_empty(self, gitlab_adapter):
        responses.add(
            responses.GET,
            "https://gitlab.com/api/v4/projects/test-owner%2Ftest-repo/protected_branches",
            json=[],
        )
        assert gitlab_adapter.list_branch_protections() == []

    @responses.activate
    def test_get_404(self, gitlab_adapter):
        responses.add(
            responses.GET,
            "https://gitlab.com/api/v4/projects/test-owner%2Ftest-repo/protected_branches/main",
            status=404,
        )
        with pytest.raises(NotFoundError):
            gitlab_adapter.get_branch_protection("main")

    @responses.activate
    def test_remove_403(self, gitlab_adapter):
        responses.add(
            responses.DELETE,
            "https://gitlab.com/api/v4/projects/test-owner%2Ftest-repo/protected_branches/main",
            status=403,
        )
        with pytest.raises(AuthenticationError):
            gitlab_adapter.remove_branch_protection("main")


# --- Gitea ---


class TestGiteaBranchProtect:
    @responses.activate
    def test_list(self, gitea_adapter):
        responses.add(
            responses.GET,
            "https://gitea.example.com/api/v1/repos/test-owner/test-repo/branch_protections",
            json=[
                {
                    "branch_name": "main",
                    "required_approvals": 1,
                    "status_check_contexts": ["ci"],
                    "enable_force_push": False,
                }
            ],
        )
        bps = gitea_adapter.list_branch_protections()
        assert len(bps) == 1
        assert bps[0].branch == "main"
        assert bps[0].require_status_checks == ("ci",)

    @responses.activate
    def test_get(self, gitea_adapter):
        responses.add(
            responses.GET,
            "https://gitea.example.com/api/v1/repos/test-owner/test-repo/branch_protections/main",
            json={
                "branch_name": "main",
                "required_approvals": 0,
                "status_check_contexts": [],
                "enable_force_push": True,
            },
        )
        bp = gitea_adapter.get_branch_protection("main")
        assert bp.allow_force_push is True

    @responses.activate
    def test_set(self, gitea_adapter):
        responses.add(
            responses.POST,
            "https://gitea.example.com/api/v1/repos/test-owner/test-repo/branch_protections",
            json={
                "branch_name": "main",
                "required_approvals": 2,
                "status_check_contexts": [],
                "enable_force_push": False,
            },
        )
        bp = gitea_adapter.set_branch_protection("main", require_reviews=2)
        assert bp.require_reviews == 2

    @responses.activate
    def test_remove(self, gitea_adapter):
        responses.add(
            responses.DELETE,
            "https://gitea.example.com/api/v1/repos/test-owner/test-repo/branch_protections/main",
            status=204,
        )
        gitea_adapter.remove_branch_protection("main")

    @responses.activate
    def test_list_empty(self, gitea_adapter):
        responses.add(
            responses.GET,
            "https://gitea.example.com/api/v1/repos/test-owner/test-repo/branch_protections",
            json=[],
        )
        assert gitea_adapter.list_branch_protections() == []

    @responses.activate
    def test_get_404(self, gitea_adapter):
        responses.add(
            responses.GET,
            "https://gitea.example.com/api/v1/repos/test-owner/test-repo/branch_protections/main",
            status=404,
        )
        with pytest.raises(NotFoundError):
            gitea_adapter.get_branch_protection("main")

    @responses.activate
    def test_remove_403(self, gitea_adapter):
        responses.add(
            responses.DELETE,
            "https://gitea.example.com/api/v1/repos/test-owner/test-repo/branch_protections/main",
            status=403,
        )
        with pytest.raises(AuthenticationError):
            gitea_adapter.remove_branch_protection("main")


# --- Forgejo ---


class TestForgejoBranchProtect:
    @responses.activate
    def test_list(self, forgejo_adapter):
        responses.add(
            responses.GET,
            "https://forgejo.example.com/api/v1/repos/test-owner/test-repo/branch_protections",
            json=[
                {
                    "branch_name": "main",
                    "required_approvals": 0,
                    "status_check_contexts": [],
                    "enable_force_push": False,
                }
            ],
        )
        bps = forgejo_adapter.list_branch_protections()
        assert len(bps) == 1

    @responses.activate
    def test_get(self, forgejo_adapter):
        responses.add(
            responses.GET,
            "https://forgejo.example.com/api/v1/repos/test-owner/test-repo/branch_protections/main",
            json={
                "branch_name": "main",
                "required_approvals": 1,
                "status_check_contexts": ["ci"],
                "enable_force_push": False,
            },
        )
        bp = forgejo_adapter.get_branch_protection("main")
        assert bp.branch == "main"
        assert bp.require_reviews == 1
        assert bp.allow_force_push is False

    @responses.activate
    def test_remove(self, forgejo_adapter):
        responses.add(
            responses.DELETE,
            "https://forgejo.example.com/api/v1/repos/test-owner/test-repo/branch_protections/main",
            status=204,
        )
        forgejo_adapter.remove_branch_protection("main")

    @responses.activate
    def test_list_empty(self, forgejo_adapter):
        responses.add(
            responses.GET,
            "https://forgejo.example.com/api/v1/repos/test-owner/test-repo/branch_protections",
            json=[],
        )
        assert forgejo_adapter.list_branch_protections() == []

    @responses.activate
    def test_get_404(self, forgejo_adapter):
        responses.add(
            responses.GET,
            "https://forgejo.example.com/api/v1/repos/test-owner/test-repo/branch_protections/main",
            status=404,
        )
        with pytest.raises(NotFoundError):
            forgejo_adapter.get_branch_protection("main")


# --- Bitbucket ---


class TestBitbucketBranchProtect:
    @responses.activate
    def test_list(self, bitbucket_adapter):
        responses.add(
            responses.GET,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions",
            json={
                "values": [
                    {"pattern": "main", "kind": "force", "id": 1},
                    {"pattern": "main", "kind": "delete", "id": 2},
                ],
                "pagelen": 10,
            },
        )
        bps = bitbucket_adapter.list_branch_protections()
        assert len(bps) == 1
        assert bps[0].branch == "main"
        assert bps[0].allow_force_push is False
        assert bps[0].allow_deletions is False

    @responses.activate
    def test_get(self, bitbucket_adapter):
        """get_branch_protection は全制限を取得し force/delete でフィルタする。"""
        responses.add(
            responses.GET,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions",
            json={
                "values": [
                    {"pattern": "main", "kind": "force", "id": 1},
                    {"pattern": "main", "kind": "delete", "id": 2},
                    {"pattern": "develop", "kind": "force", "id": 3},
                ],
                "pagelen": 10,
            },
        )
        bp = bitbucket_adapter.get_branch_protection("main")
        assert bp.branch == "main"
        assert bp.allow_force_push is False
        assert bp.allow_deletions is False

    @responses.activate
    def test_get_no_restrictions(self, bitbucket_adapter):
        """制限がないブランチでは force_push / deletions ともに許可。"""
        responses.add(
            responses.GET,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions",
            json={"values": [], "pagelen": 10},
        )
        bp = bitbucket_adapter.get_branch_protection("main")
        assert bp.allow_force_push is True
        assert bp.allow_deletions is True

    @responses.activate
    def test_set(self, bitbucket_adapter):
        """set_branch_protection は force/delete ルールを POST し結果を get で取得。"""
        # POST force
        responses.add(
            responses.POST,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions",
            json={"pattern": "main", "kind": "force", "id": 10},
            status=201,
        )
        # POST delete
        responses.add(
            responses.POST,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions",
            json={"pattern": "main", "kind": "delete", "id": 11},
            status=201,
        )
        # get_branch_protection 内で GET
        responses.add(
            responses.GET,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions",
            json={
                "values": [
                    {"pattern": "main", "kind": "force", "id": 10},
                    {"pattern": "main", "kind": "delete", "id": 11},
                ],
                "pagelen": 10,
            },
        )
        bp = bitbucket_adapter.set_branch_protection(
            "main", allow_force_push=False, allow_deletions=False
        )
        assert bp.allow_force_push is False
        assert bp.allow_deletions is False
        # POST が 2 回呼ばれたことを確認
        post_calls = [c for c in responses.calls if c.request.method == "POST"]
        assert len(post_calls) == 2

    @responses.activate
    def test_remove(self, bitbucket_adapter):
        responses.add(
            responses.GET,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions",
            json={
                "values": [{"pattern": "main", "kind": "force", "id": 1}],
                "pagelen": 10,
            },
        )
        responses.add(
            responses.DELETE,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions/1",
            status=204,
        )
        bitbucket_adapter.remove_branch_protection("main")

    @responses.activate
    def test_list_empty(self, bitbucket_adapter):
        responses.add(
            responses.GET,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions",
            json={"values": [], "pagelen": 10},
        )
        assert bitbucket_adapter.list_branch_protections() == []

    @responses.activate
    def test_list_404(self, bitbucket_adapter):
        responses.add(
            responses.GET,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions",
            status=404,
        )
        with pytest.raises(NotFoundError):
            bitbucket_adapter.list_branch_protections()

    @responses.activate
    def test_list_403(self, bitbucket_adapter):
        responses.add(
            responses.GET,
            "https://api.bitbucket.org/2.0/repositories/test-workspace/test-repo/branch-restrictions",
            status=403,
        )
        with pytest.raises(AuthenticationError):
            bitbucket_adapter.list_branch_protections()


# --- NotSupported ---


class TestBranchProtectNotSupported:
    def test_azure_devops_list(self, azure_devops_adapter):
        with pytest.raises(NotSupportedError):
            azure_devops_adapter.list_branch_protections()

    def test_azure_devops_get(self, azure_devops_adapter):
        with pytest.raises(NotSupportedError):
            azure_devops_adapter.get_branch_protection("main")

    def test_azure_devops_set(self, azure_devops_adapter):
        with pytest.raises(NotSupportedError):
            azure_devops_adapter.set_branch_protection("main")

    def test_azure_devops_remove(self, azure_devops_adapter):
        with pytest.raises(NotSupportedError):
            azure_devops_adapter.remove_branch_protection("main")

    def test_gogs_list(self, gogs_adapter):
        with pytest.raises(NotSupportedError):
            gogs_adapter.list_branch_protections()

    def test_gitbucket_list(self, gitbucket_adapter):
        with pytest.raises(NotSupportedError):
            gitbucket_adapter.list_branch_protections()

    def test_backlog_list(self, backlog_adapter):
        with pytest.raises(NotSupportedError):
            backlog_adapter.list_branch_protections()



"""gfo.commands.notification のテスト。"""

from __future__ import annotations

import json

import pytest

from gfo.adapter.base import Notification
from gfo.commands import notification as notif_cmd
from gfo.exceptions import GfoError, HttpError
from tests.test_commands.conftest import make_args, patch_adapter

SAMPLE_NOTIF = Notification(
    id="1",
    title="Fix bug",
    reason="mention",
    unread=True,
    repository="owner/repo",
    url="https://github.com/owner/repo/issues/1",
    updated_at="2024-01-01T00:00:00Z",
)


class TestHandleList:
    def test_calls_list_notifications(self, capsys):
        with patch_adapter("gfo.commands.notification") as adapter:
            adapter.list_notifications.return_value = [SAMPLE_NOTIF]
            args = make_args(unread_only=False, limit=30)
            notif_cmd.handle_list(args, fmt="table")
        adapter.list_notifications.assert_called_once_with(unread_only=False, limit=30)
        out = capsys.readouterr().out
        assert "Fix bug" in out

    def test_unread_only(self, capsys):
        with patch_adapter("gfo.commands.notification") as adapter:
            adapter.list_notifications.return_value = []
            args = make_args(unread_only=True, limit=30)
            notif_cmd.handle_list(args, fmt="table")
        adapter.list_notifications.assert_called_once_with(unread_only=True, limit=30)

    def test_json_format(self, capsys):
        with patch_adapter("gfo.commands.notification") as adapter:
            adapter.list_notifications.return_value = [SAMPLE_NOTIF]
            args = make_args(unread_only=False, limit=30)
            notif_cmd.handle_list(args, fmt="json")
        out = capsys.readouterr().out
        parsed = json.loads(out)
        assert isinstance(parsed, list)
        assert parsed[0]["title"] == "Fix bug"

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.notification") as adapter:
            adapter.list_notifications.side_effect = HttpError(500, "Server error")
            args = make_args(unread_only=False, limit=30)
            with pytest.raises(HttpError):
                notif_cmd.handle_list(args, fmt="table")


class TestHandleRead:
    def test_read_single(self):
        with patch_adapter("gfo.commands.notification") as adapter:
            args = make_args(id="1", mark_all=False)
            notif_cmd.handle_read(args, fmt="table")
        adapter.mark_notification_read.assert_called_once_with("1")

    def test_read_all(self):
        with patch_adapter("gfo.commands.notification") as adapter:
            args = make_args(id=None, mark_all=True)
            notif_cmd.handle_read(args, fmt="table")
        adapter.mark_all_notifications_read.assert_called_once()

    def test_error_both_id_and_all(self):
        with patch_adapter("gfo.commands.notification"):
            args = make_args(id="1", mark_all=True)
            with pytest.raises(GfoError, match="Cannot specify both"):
                notif_cmd.handle_read(args, fmt="table")

    def test_error_neither_id_nor_all(self):
        with patch_adapter("gfo.commands.notification"):
            args = make_args(id=None, mark_all=False)
            with pytest.raises(GfoError, match="Specify a notification"):
                notif_cmd.handle_read(args, fmt="table")

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.notification") as adapter:
            adapter.mark_notification_read.side_effect = HttpError(404, "Not found")
            args = make_args(id="999", mark_all=False)
            with pytest.raises(HttpError):
                notif_cmd.handle_read(args, fmt="table")



"""gfo.commands.gpg_key のテスト。"""

from __future__ import annotations

import json

import pytest

from gfo.adapter.base import GpgKey
from gfo.commands import gpg_key as gpg_key_cmd
from gfo.exceptions import HttpError
from tests.test_commands.conftest import make_args, patch_adapter

SAMPLE_GPG_KEY = GpgKey(
    id=1,
    primary_key_id="ABC123",
    public_key="-----BEGIN PGP PUBLIC KEY BLOCK-----...",
    emails=("user@example.com",),
    created_at="2024-01-01T00:00:00Z",
)


class TestHandleView:
    def test_calls_get_gpg_key(self):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            adapter.get_gpg_key.return_value = SAMPLE_GPG_KEY
            args = make_args(id=1)
            gpg_key_cmd.handle_view(args, fmt="table")
        adapter.get_gpg_key.assert_called_once_with(1)

    def test_outputs_result(self, capsys):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            adapter.get_gpg_key.return_value = SAMPLE_GPG_KEY
            args = make_args(id=1)
            gpg_key_cmd.handle_view(args, fmt="table")
        out = capsys.readouterr().out
        assert "ABC123" in out

    def test_json_format(self, capsys):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            adapter.get_gpg_key.return_value = SAMPLE_GPG_KEY
            args = make_args(id=1)
            gpg_key_cmd.handle_view(args, fmt="json")
        out = capsys.readouterr().out
        parsed = json.loads(out)
        assert isinstance(parsed, list)
        assert parsed[0]["primary_key_id"] == "ABC123"

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            adapter.get_gpg_key.side_effect = HttpError(404, "Not found")
            args = make_args(id=999)
            with pytest.raises(HttpError):
                gpg_key_cmd.handle_view(args, fmt="table")


class TestHandleList:
    def test_calls_list_gpg_keys(self, capsys):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            adapter.list_gpg_keys.return_value = [SAMPLE_GPG_KEY]
            args = make_args(limit=30)
            gpg_key_cmd.handle_list(args, fmt="table")
        adapter.list_gpg_keys.assert_called_once_with(limit=30)
        out = capsys.readouterr().out
        assert "ABC123" in out

    def test_json_format(self, capsys):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            adapter.list_gpg_keys.return_value = [SAMPLE_GPG_KEY]
            args = make_args(limit=30)
            gpg_key_cmd.handle_list(args, fmt="json")
        out = capsys.readouterr().out
        parsed = json.loads(out)
        assert isinstance(parsed, list)
        assert parsed[0]["primary_key_id"] == "ABC123"

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            adapter.list_gpg_keys.side_effect = HttpError(500, "Server error")
            args = make_args(limit=30)
            with pytest.raises(HttpError):
                gpg_key_cmd.handle_list(args, fmt="table")


class TestHandleCreate:
    def test_creates_gpg_key(self):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            adapter.create_gpg_key.return_value = SAMPLE_GPG_KEY
            args = make_args(key="-----BEGIN PGP PUBLIC KEY BLOCK-----...")
            gpg_key_cmd.handle_create(args, fmt="table")
        adapter.create_gpg_key.assert_called_once_with(
            armored_key="-----BEGIN PGP PUBLIC KEY BLOCK-----..."
        )

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            adapter.create_gpg_key.side_effect = HttpError(422, "Unprocessable")
            args = make_args(key="-----BEGIN PGP PUBLIC KEY BLOCK-----...")
            with pytest.raises(HttpError):
                gpg_key_cmd.handle_create(args, fmt="table")


class TestHandleDelete:
    def test_calls_delete_gpg_key(self, capsys):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            args = make_args(id=1)
            gpg_key_cmd.handle_delete(args, fmt="table")
        adapter.delete_gpg_key.assert_called_once_with(key_id=1)
        out = capsys.readouterr().out
        assert "Deleted" in out
        assert "1" in out

    def test_calls_delete_gpg_key_string_id(self):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            args = make_args(id="abc-123")
            gpg_key_cmd.handle_delete(args, fmt="table")
        adapter.delete_gpg_key.assert_called_once_with(key_id="abc-123")

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.gpg_key") as adapter:
            adapter.delete_gpg_key.side_effect = HttpError(404, "Not found")
            args = make_args(id=1)
            with pytest.raises(HttpError):
                gpg_key_cmd.handle_delete(args, fmt="table")



"""gfo pr comment / gfo issue comment サブコマンドのハンドラ。"""

from __future__ import annotations

import argparse

from gfo.commands import get_adapter
from gfo.exceptions import ConfigError
from gfo.i18n import _
from gfo.output import output


def _dispatch(args: argparse.Namespace, resource: str, *, fmt: str, jq: str | None = None) -> None:
    """comment サブコマンドの共通ディスパッチ。"""
    action = getattr(args, "comment_action", None)
    if action is None:
        raise ConfigError(_("comment action required: list, create, edit, delete"))
    adapter = get_adapter()
    if action == "list":
        comments = adapter.list_comments(resource, args.number, limit=args.limit)
        output(comments, fmt=fmt, fields=["id", "author", "body", "created_at"], jq=jq)
    elif action == "create":
        comment = adapter.create_comment(resource, args.number, body=args.body)
        output(comment, fmt=fmt, jq=jq)
    elif action == "edit":
        comment = adapter.update_comment(resource, args.comment_id, body=args.body)
        output(comment, fmt=fmt, jq=jq)
    elif action == "delete":
        adapter.delete_comment(resource, args.comment_id)
        print(_("Deleted comment '{comment_id}'.").format(comment_id=args.comment_id))


def handle_pr_comment(args: argparse.Namespace, *, fmt: str, jq: str | None = None) -> None:
    """gfo pr comment のハンドラ。"""
    _dispatch(args, "pr", fmt=fmt, jq=jq)


def handle_issue_comment(args: argparse.Namespace, *, fmt: str, jq: str | None = None) -> None:
    """gfo issue comment のハンドラ。"""
    _dispatch(args, "issue", fmt=fmt, jq=jq)



"""Gitea 統合テスト。

Gitea/Forgejo 共通のテストは base_gitea_family.GiteaFamilyIntegrationBase に集約。
このファイルには Gitea 固有の差分（wiki の main→master 同期など）のみ記述する。
"""

from __future__ import annotations

import os
import subprocess
import sys
import tempfile

import pytest

from tests.integration.base_gitea_family import GiteaFamilyIntegrationBase
from tests.integration.conftest import get_service_config, safe_temporary_directory

CONFIG = get_service_config("gitea")

pytestmark = [
    pytest.mark.integration,
    pytest.mark.selfhosted,
    pytest.mark.skipif(CONFIG is None, reason="Gitea credentials not configured"),
]


class TestGiteaIntegration(GiteaFamilyIntegrationBase):
    """Gitea に対する統合テスト。

    テストメソッドはアルファベット順（名前順）に実行される。
    テスト間で作成したリソースの番号をクラス変数で共有する。
    """

    CONFIG = CONFIG

    def test_01_repo_view(self) -> None:
        repo = self.adapter.get_repository()
        assert repo.name == self.config.repo
        assert self.config.owner in repo.full_name

    @classmethod
    def _make_askpass_env(cls, token: str, username: str = "gfo-admin") -> dict:
        """GIT_ASKPASS 用の環境変数辞書を生成する。

        トークンを URL に直接埋め込む代わりに、GIT_ASKPASS スクリプト経由で
        認証情報を渡す。一時ファイルにヘルパースクリプトを書き出し、
        GIT_ASKPASS に設定する。
        """
        askpass_script = os.path.join(os.path.dirname(__file__), "git_askpass_helper.py")
        # GIT_ASKPASS は実行可能ファイルのパスのみ受け付ける（引数不可）ため、
        # Python インタープリタ経由で呼び出すラッパーを一時ファイルに書き出す
        wrapper_dir = tempfile.mkdtemp(prefix="gfo-askpass-")
        if sys.platform == "win32":
            wrapper_path = os.path.join(wrapper_dir, "askpass.bat")
            with open(wrapper_path, "w") as f:
                f.write(f'@"{sys.executable}" "{askpass_script}" %1\n')
        else:
            wrapper_path = os.path.join(wrapper_dir, "askpass.sh")
            with open(wrapper_path, "w") as f:
                f.write(f'#!/bin/sh\n"{sys.executable}" "{askpass_script}" "$1"\n')
            os.chmod(wrapper_path, 0o755)
        return {
            **os.environ,
            "GIT_ASKPASS": wrapper_path,
            "GFO_GIT_USERNAME": username,
            "GFO_GIT_TOKEN": token,
            # GCM がブラウザを開かないようにする（GIT_ASKPASS を優先させる）
            "GCM_INTERACTIVE": "never",
            "GIT_TERMINAL_PROMPT": "0",
        }

    @classmethod
    def _sync_wiki_master(cls) -> None:
        """Gitea 1.22 の wiki ブランチ不整合修正: main → master を同期する。

        Gitea 1.22 は wiki への書き込みを main ブランチ、
        読み取りを master ブランチで行うため、write 後に手動で同期が必要。
        GIT_ASKPASS を使用してトークンを安全に渡す。
        """
        host = cls.config.host or "localhost:3000"
        owner = cls.config.owner
        repo = cls.config.repo
        token = cls.config.token
        wiki_url = f"http://{host}/{owner}/{repo}.wiki.git"
        git_env = cls._make_askpass_env(token)

        with safe_temporary_directory() as tmpdir:
            wiki_dir = f"{tmpdir}/wiki"
            try:
                r = subprocess.run(
                    ["git", "clone", "--depth=1", wiki_url, wiki_dir],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    stdin=subprocess.DEVNULL,
                    env=git_env,
                    timeout=30,
                )
                if r.returncode != 0:
                    return  # wiki がまだ初期化されていない場合はスキップ
                # 現在の HEAD (= main の先端) を master ブランチとして force push
                subprocess.run(
                    ["git", "push", "origin", "HEAD:refs/heads/master", "--force"],
                    cwd=wiki_dir,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    stdin=subprocess.DEVNULL,
                    env=git_env,
                    timeout=30,
                )
            except Exception:
                pass  # ベストエフォート

    # --- wiki CRUD ---

    def test_45_wiki_crud(self) -> None:
        """Wiki ページの作成・取得・一覧・更新・削除テスト。

        Gitea 1.22 固有の挙動:
        - 書き込み（create/update）は main ブランチへのコミットが成功するが、
          レスポンスは master から読もうとして 404 を返す（バグ）。
        - 読み取り（list/get）は master ブランチから行う。
        - _sync_wiki_master() で master = main に同期してから読み取る。
        - ページ識別子は sub_url（例: "gfo-test-wiki.-"）を使用する。
        """
        # Wiki を有効化
        # TODO: _client はプライベートメンバーへの依存。公開 API への移行を検討。
        try:
            self.adapter._client.patch(
                f"{self.adapter._repos_path()}",
                json={"has_wiki": True},
            )
        except Exception:
            pass
        # master を main に同期してから読み取り可能な状態にする
        self._sync_wiki_master()
        # 残留ページを削除する（sub_url で識別）
        try:
            for p in self.adapter.list_wiki_pages():
                if p.title == "gfo-test-wiki":
                    self.adapter.delete_wiki_page(p.id)
            self._sync_wiki_master()
        except Exception:
            pass
        # Create: Gitea 1.22 はページを main に作成するが 404 を返す（バグ）
        from gfo.exceptions import NotFoundError

        try:
            self.adapter.create_wiki_page(title="gfo-test-wiki", content="hello wiki content")
        except NotFoundError:
            pass  # Gitea 1.22 bug: ページは作成されているが 404 が返る
        # 同期してから読み取る
        self._sync_wiki_master()
        pages = self.adapter.list_wiki_pages()
        wiki_page = next((p for p in pages if p.title == "gfo-test-wiki"), None)
        assert wiki_page is not None, "gfo-test-wiki が作成されているはず"
        # sub_url で取得
        page_read = self.adapter.get_wiki_page(wiki_page.id)
        assert page_read.title == "gfo-test-wiki"
        assert page_read.content == "hello wiki content"
        # Update: Gitea 1.22 は master に page が存在する状態なら PATCH 200 が返る
        updated_page = self.adapter.update_wiki_page(wiki_page.id, content="updated wiki content")
        assert "updated" in updated_page.content
        # 削除して確認
        self.adapter.delete_wiki_page(wiki_page.id)
        self._sync_wiki_master()
        pages_after = self.adapter.list_wiki_pages()
        assert not any(p.title == "gfo-test-wiki" for p in pages_after)



"""gfo api コマンドのハンドラ。アダプター層をバイパスして任意の API パスにリクエストを送る。"""

from __future__ import annotations

import argparse
import json

from gfo.adapter.registry import create_http_client
from gfo.auth import resolve_token
from gfo.config import resolve_project_config
from gfo.exceptions import ConfigError
from gfo.i18n import _
from gfo.output import apply_jq_filter


def _parse_headers(header_list: list[str] | None) -> dict[str, str]:
    """--header 引数をパースして dict に変換する。"""
    if not header_list:
        return {}
    headers: dict[str, str] = {}
    for h in header_list:
        if ":" not in h:
            raise ConfigError(_("Invalid header format: '{h}'. Use 'Key: Value'.").format(h=h))
        key, value = h.split(":", 1)
        headers[key.strip()] = value.strip()
    return headers


def handle_api(args: argparse.Namespace, *, fmt: str, jq: str | None = None) -> None:
    """gfo api のハンドラ。"""
    config = resolve_project_config()
    token = resolve_token(config.host, config.service_type)
    client = create_http_client(config.service_type, config.api_url, token)

    method = args.method.upper()
    path = args.path
    headers = _parse_headers(getattr(args, "header", None))
    data = getattr(args, "data", None)
    try:
        json_data = json.loads(data) if data else None
    except json.JSONDecodeError as e:
        raise ConfigError(_("Invalid JSON in --data: {err}").format(err=e)) from e

    resp = client.request(method, path, json=json_data, headers=headers)
    body = resp.text

    if jq:
        print(apply_jq_filter(body, jq))
    else:
        print(body)



"""GitHub 統合テスト。"""

from __future__ import annotations

import time

import pytest

from gfo.exceptions import GfoError, NotSupportedError
from tests.integration.conftest import (
    TEST_SSH_PUBLIC_KEY,
    ServiceTestConfig,
    create_test_adapter,
    get_service_config,
    safe_temporary_directory,
)

CONFIG = get_service_config("github")

pytestmark = [
    pytest.mark.integration,
    pytest.mark.saas,
    pytest.mark.skipif(CONFIG is None, reason="GitHub credentials not configured"),
]


class TestGitHubIntegration:
    """GitHub に対する統合テスト。"""

    @classmethod
    def setup_class(cls) -> None:
        assert CONFIG is not None
        cls.adapter = create_test_adapter(CONFIG)
        cls.config = CONFIG
        cls._issue_number: int | None = None
        cls._pr_number: int | None = None
        cls._update_issue_number: int | None = None
        cls._update_pr_number: int | None = None
        cls._update_issue_comment_id: int | None = None
        cls._update_pr_comment_id: int | None = None
        cls._webhook_id: int | None = None
        cls._deploy_key_id: int | None = None
        cls._head_sha: str | None = None

    @classmethod
    def teardown_class(cls) -> None:
        """テスト終了後にテスト用リソースを削除する。"""
        try:
            cls.adapter.delete_release(tag="v0.0.1-test")
        except Exception:
            pass
        try:
            # TODO: _client, _repos_path() はプライベートメンバーへの依存。公開 API への移行を検討。
            # リリース削除では git タグが残るため個別削除
            cls.adapter._client.delete(f"{cls.adapter._repos_path()}/git/refs/tags/v0.0.1-test")
        except Exception:
            pass
        try:
            cls.adapter.delete_label(name="gfo-test-label")
        except Exception:
            pass
        try:
            for ms in cls.adapter.list_milestones():
                if ms.title == "gfo-test-milestone":
                    cls.adapter.delete_milestone(number=ms.number)
                    break
        except Exception:
            pass
        try:
            cls.adapter.delete_branch(name="gfo-test-branch-temp")
        except Exception:
            pass
        try:
            cls.adapter._client.delete(f"{cls.adapter._repos_path()}/git/refs/tags/v0.0.2-test")
        except Exception:
            pass
        try:
            if cls._webhook_id is not None:
                cls.adapter.delete_webhook(hook_id=cls._webhook_id)
        except Exception:
            pass
        try:
            if cls._deploy_key_id is not None:
                cls.adapter.delete_deploy_key(key_id=cls._deploy_key_id)
        except Exception:
            pass
        try:
            if cls._update_issue_number is not None:
                cls.adapter.close_issue(cls._update_issue_number)
        except Exception:
            pass
        try:
            if cls._update_pr_number is not None:
                cls.adapter.close_pull_request(cls._update_pr_number)
        except Exception:
            pass
        try:
            _, sha = cls.adapter.get_file_content("gfo-test-file.txt", ref=cls.config.test_branch)
            if sha:
                cls.adapter.delete_file(
                    "gfo-test-file.txt",
                    sha=sha,
                    message="cleanup: remove gfo-test-file.txt",
                    branch=cls.config.test_branch,
                )
        except Exception:
            pass

    # --- Repository ---

    def test_01_repo_view(self) -> None:
        repo = self.adapter.get_repository()
        assert repo.name == self.config.repo
        assert repo.full_name == f"{self.config.owner}/{self.config.repo}"

    def test_02_repo_list(self) -> None:
        # owner 指定は公開リポジトリのみ返すため、認証済み一覧（プライベート含む）を使用
        repos = self.adapter.list_repositories(limit=10)
        assert len(repos) > 0
        names = [r.name for r in repos]
        assert self.config.repo in names

    # --- Label ---

    def test_03_label_create(self) -> None:
        label = self.adapter.create_label(
            name="gfo-test-label",
            color="ff0000",
            description="Integration test",
        )
        assert label.name == "gfo-test-label"

    def test_04_label_list(self) -> None:
        labels = self.adapter.list_labels()
        names = [lb.name for lb in labels]
        assert "gfo-test-label" in names

    # --- Milestone ---

    def test_05_milestone_create(self) -> None:
        ms = self.adapter.create_milestone(
            title="gfo-test-milestone",
            description="Integration test",
        )
        assert ms.title == "gfo-test-milestone"

    def test_06_milestone_list(self) -> None:
        milestones = self.adapter.list_milestones()
        titles = [m.title for m in milestones]
        assert "gfo-test-milestone" in titles

    # --- Issue ---

    def test_07_issue_create(self) -> None:
        issue = self.adapter.create_issue(title="gfo-test-issue", body="Integration test")
        assert issue.title == "gfo-test-issue"
        assert issue.state == "open"
        self.__class__._issue_number = issue.number

    def test_08_issue_list(self) -> None:
        import time

        assert self._issue_number is not None
        # GitHub API は private repo で作成直後の issue がリストに反映されるまで遅延がある場合がある
        time.sleep(2)
        numbers: list[int] = []
        for _ in range(5):
            issues = self.adapter.list_issues(state="open", limit=0)
            numbers = [i.number for i in issues]
            if self._issue_number in numbers:
                break
            time.sleep(2)
        assert self._issue_number in numbers

    def test_09_issue_view(self) -> None:
        assert self._issue_number is not None
        issue = self.adapter.get_issue(self._issue_number)
        assert issue.title == "gfo-test-issue"

    def test_10_issue_close(self) -> None:
        assert self._issue_number is not None
        self.adapter.close_issue(self._issue_number)
        issue = self.adapter.get_issue(self._issue_number)
        assert issue.state == "closed"

    # --- Pull Request ---

    def test_11_pr_create(self) -> None:
        import base64
        import time

        from gfo.exceptions import GfoError

        # TODO: _repos_path(), _client はプライベートメンバーへの依存。公開 API への移行を検討。
        # 前回マージ済みの場合はブランチに差分がないため、テストファイルを更新してコミットを追加する
        content = base64.b64encode(f"test run {time.time()}".encode()).decode()
        marker_path = f"{self.adapter._repos_path()}/contents/test-pr-marker.txt"
        payload: dict = {
            "message": "test: update marker for PR",
            "content": content,
            "branch": self.config.test_branch,
        }
        try:
            existing = self.adapter._client.get(
                marker_path, params={"ref": self.config.test_branch}
            )
            payload["sha"] = existing.json()["sha"]
        except GfoError:
            pass  # ファイルが存在しない場合は sha なしで新規作成
        self.adapter._client.put(marker_path, json=payload)

        pr = self.adapter.create_pull_request(
            title="gfo-test-pr",
            body="Integration test",
            base=self.config.default_branch,
            head=self.config.test_branch,
        )
        assert pr.title == "gfo-test-pr"
        assert pr.state == "open"
        self.__class__._pr_number = pr.number

    def test_12_pr_list(self) -> None:
        assert self._pr_number is not None
        prs = self.adapter.list_pull_requests(state="open", limit=10)
        numbers = [p.number for p in prs]
        assert self._pr_number in numbers

    def test_13_pr_view(self) -> None:
        assert self._pr_number is not None
        pr = self.adapter.get_pull_request(self._pr_number)
        assert pr.title == "gfo-test-pr"

    def test_14_pr_merge(self) -> None:
        assert self._pr_number is not None
        self.adapter.merge_pull_request(self._pr_number, method="merge")
        pr = self.adapter.get_pull_request(self._pr_number)
        assert pr.state == "merged"

    # --- Release ---

    def test_15_release_create(self) -> None:
        release = self.adapter.create_release(
            tag="v0.0.1-test",
            title="Test Release",
            notes="Integration test",
        )
        assert release.tag == "v0.0.1-test"

    def test_16_release_list(self) -> None:
        import time

        # GitHub API は作成直後のリリースがリストに反映されるまで遅延がある場合がある
        tags: list[str] = []
        for _ in range(5):
            releases = self.adapter.list_releases(limit=0)
            tags = [r.tag for r in releases]
            if "v0.0.1-test" in tags:
                break
            time.sleep(2)
        assert "v0.0.1-test" in tags

    # --- PR close ---

    def test_17_pr_close(self) -> None:
        import base64
        import time

        # TODO: _repos_path(), _client はプライベートメンバーへの依存。公開 API への移行を検討。
        content = base64.b64encode(f"close-test {time.time()}".encode()).decode()
        marker_path = f"{self.adapter._repos_path()}/contents/test-close-marker.txt"
        payload: dict = {
            "message": "test: add marker for close test",
            "content": content,
            "branch": self.config.test_branch,
        }
        try:
            existing = self.adapter._client.get(
                marker_path, params={"ref": self.config.test_branch}
            )
            payload["sha"] = existing.json()["sha"]
        except GfoError:
            pass
        self.adapter._client.put(marker_path, json=payload)
        pr = self.adapter.create_pull_request(
            title="gfo-test-pr-close",
            body="Integration test for close",
            base=self.config.default_branch,
            head=self.config.test_branch,
        )
        self.adapter.close_pull_request(pr.number)
        closed = self.adapter.get_pull_request(pr.number)
        assert closed.state == "closed"

    # --- Issue delete ---

    def test_18_issue_delete_not_supported(self) -> None:
        assert self._issue_number is not None
        with pytest.raises(NotSupportedError):
            self.adapter.delete_issue(self._issue_number)

    # --- Label delete ---

    def test_19_label_delete(self) -> None:
        self.adapter.delete_label(name="gfo-test-label")
        labels = self.adapter.list_labels()
        assert not any(lb.name == "gfo-test-label" for lb in labels)

    # --- Milestone delete ---

    def test_20_milestone_delete(self) -> None:
        milestones = self.adapter.list_milestones()
        test_ms = next((m for m in milestones if m.title == "gfo-test-milestone"), None)
        assert test_ms is not None
        self.adapter.delete_milestone(number=test_ms.number)
        milestones_after = self.adapter.list_milestones()
        assert not any(m.title == "gfo-test-milestone" for m in milestones_after)

    # --- Release delete ---

    def test_21_release_delete(self) -> None:
        self.adapter.delete_release(tag="v0.0.1-test")
        releases = self.adapter.list_releases()
        assert not any(r.tag == "v0.0.1-test" for r in releases)

    # --- Repo create and delete ---

    def test_22_repo_create_and_delete(self) -> None:
        import time

        temp_name = f"gfo-test-temp-{int(time.time())}"
        temp_adapter = None
        try:
            repo = self.adapter.create_repository(
                name=temp_name, private=True, description="Integration test temp"
            )
            assert repo.name == temp_name
            temp_config = ServiceTestConfig(
                service_type=self.config.service_type,
                host=self.config.host,
                api_url=self.config.api_url,
                owner=self.config.owner,
                repo=temp_name,
                token=self.config.token,
                organization=self.config.organization,
                project_key=self.config.project_key,
            )
            temp_adapter = create_test_adapter(temp_config)
            temp_adapter.delete_repository()
        except Exception:
            if temp_adapter is not None:
                try:
                    temp_adapter.delete_repository()
                except Exception:
                    pass
            raise

    # --- update_issue ---

    def test_23_update_issue(self) -> None:
        """Issue の title/body 更新テスト。"""
        issue = self.adapter.create_issue(
            title="gfo-test-update-issue",
            body="original body",
        )
        self.__class__._update_issue_number = issue.number
        updated = self.adapter.update_issue(
            issue.number,
            title="gfo-test-update-issue-updated",
            body="updated body",
        )
        assert updated.title == "gfo-test-update-issue-updated"

    # --- update_pr ---

    def test_24_update_pr(self) -> None:
        """PR の title 更新テスト。差分確保のため test_branch にコミットを追加してから PR 作成。"""
        import base64
        import time

        # TODO: _repos_path(), _client はプライベートメンバーへの依存。公開 API への移行を検討。
        content = base64.b64encode(f"update-pr-{time.time()}".encode()).decode()
        marker_path = f"{self.adapter._repos_path()}/contents/test-update-pr-marker.txt"
        payload: dict = {
            "message": "test: add marker for update PR",
            "content": content,
            "branch": self.config.test_branch,
        }
        try:
            existing = self.adapter._client.get(
                marker_path, params={"ref": self.config.test_branch}
            )
            payload["sha"] = existing.json()["sha"]
        except GfoError:
            pass
        self.adapter._client.put(marker_path, json=payload)
        pr = self.adapter.create_pull_request(
            title="gfo-test-update-pr",
            body="original body",
            base=self.config.default_branch,
            head=self.config.test_branch,
        )
        self.__class__._update_pr_number = pr.number
        updated_pr = self.adapter.update_pull_request(
            pr.number,
            title="gfo-test-update-pr-updated",
        )
        assert updated_pr.title == "gfo-test-update-pr-updated"

    # --- create_comment (issue) ---

    def test_25_create_issue_comment(self) -> None:
        """Issue にコメントを作成するテスト。"""
        assert self._update_issue_number is not None
        comment = self.adapter.create_comment(
            "issue", self._update_issue_number, body="test comment body"
        )
        assert comment.body == "test comment body"
        self.__class__._update_issue_comment_id = comment.id

    # --- list_comments (issue) ---

    def test_26_list_issue_comments(self) -> None:
        """Issue コメント一覧に test_25 で作成したコメントが含まれることを確認する。"""
        assert self._update_issue_number is not None
        assert self._update_issue_comment_id is not None
        comments = self.adapter.list_comments("issue", self._update_issue_number)
        assert any(c.id == self._update_issue_comment_id for c in comments)

    # --- update_comment ---

    def test_27_update_comment(self) -> None:
        """コメント本文を更新するテスト。"""
        assert self._update_issue_comment_id is not None
        updated = self.adapter.update_comment(
            "issue", self._update_issue_comment_id, body="updated comment body"
        )
        assert updated.body == "updated comment body"

    # --- delete_comment ---

    def test_28_delete_comment(self) -> None:
        """コメントを削除するテスト。"""
        assert self._update_issue_comment_id is not None
        assert self._update_issue_number is not None
        self.adapter.delete_comment("issue", self._update_issue_comment_id)
        comments = self.adapter.list_comments("issue", self._update_issue_number)
        assert not any(c.id == self._update_issue_comment_id for c in comments)

    # --- PR comment ---

    def test_29_pr_comment(self) -> None:
        """PR にコメントを作成・一覧取得するテスト。"""
        assert self._update_pr_number is not None
        comment = self.adapter.create_comment("pr", self._update_pr_number, body="test PR comment")
        assert comment.body == "test PR comment"
        self.__class__._update_pr_comment_id = comment.id
        comments = self.adapter.list_comments("pr", self._update_pr_number)
        assert any(c.id == self._update_pr_comment_id for c in comments)

    # --- review ---

    def test_30_review(self) -> None:
        """PR にレビューを作成・一覧取得するテスト。"""
        assert self._update_pr_number is not None
        review = self.adapter.create_review(
            self._update_pr_number, state="COMMENT", body="test review"
        )
        assert review.body == "test review"
        reviews = self.adapter.list_reviews(self._update_pr_number)
        assert len(reviews) > 0

    # test_31 は欠番: 全サービス共通で test_30 (review) と test_32 (list_branches) の間に
    # 予約されていた番号だが、該当する機能テストが不要になったためスキップされている。

    # --- list_branches ---

    def test_32_list_branches(self) -> None:
        """ブランチ一覧にデフォルトブランチが含まれることを確認する。"""
        branches = self.adapter.list_branches()
        names = [b.name for b in branches]
        assert self.config.default_branch in names

    # --- create_branch ---

    def test_33_create_branch(self) -> None:
        """テスト用ブランチを作成するテスト。"""
        branch = self.adapter.create_branch(
            name="gfo-test-branch-temp",
            ref=self.config.default_branch,
        )
        assert branch.name == "gfo-test-branch-temp"

    # --- delete_branch ---

    def test_34_delete_branch(self) -> None:
        """test_33 で作成したブランチを削除するテスト。"""
        self.adapter.delete_branch(name="gfo-test-branch-temp")
        branches = self.adapter.list_branches()
        assert not any(b.name == "gfo-test-branch-temp" for b in branches)

    # --- create_tag ---

    def test_35_create_tag(self) -> None:
        """タグを作成するテスト。"""
        # TODO: _client, _repos_path() はプライベートメンバーへの依存。公開 API への移行を検討。
        branch_resp = self.adapter._client.get(
            f"{self.adapter._repos_path()}/branches/{self.config.default_branch}"
        )
        head_sha = branch_resp.json()["commit"]["sha"]
        tag = self.adapter.create_tag(name="v0.0.2-test", ref=head_sha)
        assert tag.name == "v0.0.2-test"

    # --- list_tags ---

    def test_36_list_tags(self) -> None:
        """タグ一覧に test_35 で作成したタグが含まれることを確認する。"""
        tags = self.adapter.list_tags()
        assert any(t.name == "v0.0.2-test" for t in tags)

    # --- delete_tag ---

    def test_37_delete_tag(self) -> None:
        """test_35 で作成したタグを削除するテスト。"""
        self.adapter.delete_tag(name="v0.0.2-test")
        # GitHub はリリースとタグ ref を別管理するため、タグ ref も削除する
        try:
            self.adapter._client.delete(f"{self.adapter._repos_path()}/git/refs/tags/v0.0.2-test")
        except Exception:
            pass
        tags = self.adapter.list_tags()
        assert not any(t.name == "v0.0.2-test" for t in tags)

    # --- create_commit_status ---

    def test_38_create_commit_status(self) -> None:
        """コミットステータスを作成するテスト。"""
        # TODO: _client, _repos_path() はプライベートメンバーへの依存。公開 API への移行を検討。
        branch_resp = self.adapter._client.get(
            f"{self.adapter._repos_path()}/branches/{self.config.default_branch}"
        )
        self.__class__._head_sha = branch_resp.json()["commit"]["sha"]
        status = self.adapter.create_commit_status(
            self._head_sha,
            state="success",
            context="gfo-ci/test",
            description="Integration test status",
        )
        assert status.state == "success"

    # --- list_commit_statuses ---

    def test_39_list_commit_statuses(self) -> None:
        """コミットステータス一覧を取得するテスト。"""
        assert self._head_sha is not None
        statuses = self.adapter.list_commit_statuses(self._head_sha)
        assert isinstance(statuses, list)

    # --- file CRUD ---

    def test_40_file_crud(self) -> None:
        """ファイルの作成・取得・更新・削除テスト。"""
        # 前回テストで残ったファイルをクリーンアップ
        try:
            _, sha = self.adapter.get_file_content("gfo-test-file.txt", ref=self.config.test_branch)
            if sha:
                self.adapter.delete_file(
                    "gfo-test-file.txt",
                    sha=sha,
                    message="cleanup: remove leftover gfo-test-file.txt",
                    branch=self.config.test_branch,
                )
        except Exception:
            pass
        self.adapter.create_or_update_file(
            "gfo-test-file.txt",
            content="hello gfo",
            message="test: create gfo-test-file.txt",
            branch=self.config.test_branch,
        )
        content, sha = self.adapter.get_file_content(
            "gfo-test-file.txt", ref=self.config.test_branch
        )
        assert content == "hello gfo"
        commit_sha = self.adapter.create_or_update_file(
            "gfo-test-file.txt",
            content="updated gfo",
            message="test: update gfo-test-file.txt",
            sha=sha,
            branch=self.config.test_branch,
        )
        # commit SHA で参照することで GitHub のブランチキャッシュ遅延を回避する
        ref = commit_sha if commit_sha else self.config.test_branch
        content2, sha2 = self.adapter.get_file_content("gfo-test-file.txt", ref=ref)
        assert content2 == "updated gfo"
        self.adapter.delete_file(
            "gfo-test-file.txt",
            sha=sha2,
            message="test: delete gfo-test-file.txt",
            branch=self.config.test_branch,
        )

    # --- webhook CRUD ---

    def test_41_webhook_crud(self) -> None:
        """Webhook の作成・一覧・削除テスト。"""
        try:
            for h in self.adapter.list_webhooks():
                if h.url == "https://example.com/webhook":
                    self.adapter.delete_webhook(hook_id=h.id)
        except Exception:
            pass
        hook = self.adapter.create_webhook(
            url="https://example.com/webhook",
            events=["push"],
        )
        assert hook.url == "https://example.com/webhook"
        self.__class__._webhook_id = hook.id
        hooks = self.adapter.list_webhooks()
        assert any(h.id == self._webhook_id for h in hooks)
        self.adapter.delete_webhook(hook_id=self._webhook_id)
        hooks_after = self.adapter.list_webhooks()
        assert not any(h.id == self._webhook_id for h in hooks_after)

    # --- deploy_key CRUD ---

    def test_42_deploy_key_crud(self) -> None:
        """デプロイキーの作成・一覧・削除テスト。"""
        try:
            for k in self.adapter.list_deploy_keys():
                if k.title == "gfo-test-deploy-key":
                    self.adapter.delete_deploy_key(key_id=k.id)
        except Exception:
            pass
        key = self.adapter.create_deploy_key(
            title="gfo-test-deploy-key",
            key=TEST_SSH_PUBLIC_KEY,
        )
        assert key.title == "gfo-test-deploy-key"
        self.__class__._deploy_key_id = key.id
        keys = self.adapter.list_deploy_keys()
        assert any(k.id == self._deploy_key_id for k in keys)
        self.adapter.delete_deploy_key(key_id=self._deploy_key_id)
        keys_after = self.adapter.list_deploy_keys()
        assert not any(k.id == self._deploy_key_id for k in keys_after)

    # --- get_current_user ---

    def test_43_get_current_user(self) -> None:
        """現在のユーザー情報を取得するテスト。"""
        user = self.adapter.get_current_user()
        assert isinstance(user, dict)
        assert "login" in user

    # --- search + misc（wiki NSE 含む）---

    def test_44_search_and_misc(self) -> None:
        """search, list_collaborators, get_pr_checkout_refspec, wiki NSE テスト。"""
        repos = self.adapter.search_repositories(self.config.repo[:4], limit=5)
        assert isinstance(repos, list)
        issues = self.adapter.search_issues("gfo-test", limit=5)
        assert isinstance(issues, list)
        collaborators = self.adapter.list_collaborators()
        assert isinstance(collaborators, list)
        assert self._pr_number is not None
        refspec = self.adapter.get_pr_checkout_refspec(self._pr_number)
        assert refspec
        # GitHub は wiki API 非対応
        with pytest.raises(NotSupportedError):
            self.adapter.list_wiki_pages()
        with pytest.raises(NotSupportedError):
            self.adapter.create_wiki_page(title="test", content="test")

    # --- browse ---

    def test_45_browse(self) -> None:
        """get_web_url で Web URL を取得するテスト（--print モード相当）。"""
        url = self.adapter.get_web_url()
        assert isinstance(url, str)
        assert len(url) > 0
        assert "github.com" in url

    # --- ssh-key CRUD ---

    def test_46_ssh_key_crud(self) -> None:
        """SSH キーの作成・一覧・削除テスト。"""
        import os
        import subprocess

        with safe_temporary_directory() as tmpdir:
            key_path = os.path.join(tmpdir, "gfo_test_key")
            subprocess.run(
                ["ssh-keygen", "-t", "ed25519", "-f", key_path, "-N", "", "-C", "gfo-test"],
                capture_output=True,
                check=True,
            )
            with open(key_path + ".pub") as f:
                dummy_key = f.read().strip()
        # 残留キーをクリーンアップ
        try:
            for k in self.adapter.list_ssh_keys():
                if k.title == "gfo-test-ssh-key":
                    self.adapter.delete_ssh_key(key_id=k.id)
        except Exception:
            pass
        key = self.adapter.create_ssh_key(title="gfo-test-ssh-key", key=dummy_key)
        assert key.title == "gfo-test-ssh-key"
        keys = self.adapter.list_ssh_keys()
        assert any(k.id == key.id for k in keys)
        self.adapter.delete_ssh_key(key_id=key.id)
        keys_after = self.adapter.list_ssh_keys()
        assert not any(k.id == key.id for k in keys_after)

    # --- org ---

    def test_47_org_list(self) -> None:
        """Organization 一覧取得テスト。"""
        orgs = self.adapter.list_organizations()
        assert isinstance(orgs, list)

    def test_47b_org_get(self) -> None:
        """Organization 詳細取得テスト。"""
        orgs = self.adapter.list_organizations(limit=1)
        if not orgs:
            pytest.skip("No organizations available for testing")
        org = self.adapter.get_organization(orgs[0].name)
        assert org.name == orgs[0].name

    def test_47c_org_members(self) -> None:
        """Organization メンバー一覧テスト。"""
        orgs = self.adapter.list_organizations(limit=1)
        if not orgs:
            pytest.skip("No organizations available for testing")
        members = self.adapter.list_org_members(orgs[0].name)
        assert isinstance(members, list)

    def test_47d_org_repos(self) -> None:
        """Organization のリポジトリ一覧テスト。"""
        orgs = self.adapter.list_organizations(limit=1)
        if not orgs:
            pytest.skip("No organizations available for testing")
        repos = self.adapter.list_repositories(owner=orgs[0].name, limit=5)
        assert isinstance(repos, list)

    # --- notification ---

    def test_48_notification_list(self) -> None:
        """通知一覧取得テスト。classic token が必要な場合があるため try/except でスキップ。"""
        try:
            notifications = self.adapter.list_notifications(limit=5)
            assert isinstance(notifications, list)
        except Exception:
            pytest.skip("Notification API requires classic token with notifications scope")

    def test_48b_notification_mark_read(self) -> None:
        """通知の既読マークテスト。通知がない場合は mark_all_notifications_read のみ確認。"""
        try:
            # mark_all_notifications_read は通知がなくても成功する
            self.adapter.mark_all_notifications_read()
        except Exception:
            pytest.skip("Notification API requires classic token with notifications scope")

    # --- branch-protect ---

    def test_49_branch_protect_list(self) -> None:
        """ブランチ保護の一覧取得テスト。空リストでも OK。"""
        protections = self.adapter.list_branch_protections()
        assert isinstance(protections, list)

    # --- secret ---

    def test_50_secret_crud(self) -> None:
        """Secret の set → list → delete テスト。PyNaCl が必要なためスキップ可能。"""
        try:
            from nacl import public  # noqa: F401
        except ImportError:
            pytest.skip("PyNaCl is required for GitHub secret encryption")
        # クリーンアップ
        try:
            self.adapter.delete_secret("GFO_TEST_SECRET")
        except Exception:
            pass
        secret = self.adapter.set_secret("GFO_TEST_SECRET", "test-value")
        assert secret.name == "GFO_TEST_SECRET"
        time.sleep(3)  # GitHub Actions secrets API の反映ラグ対策
        secrets = self.adapter.list_secrets()
        assert any(s.name == "GFO_TEST_SECRET" for s in secrets)
        self.adapter.delete_secret("GFO_TEST_SECRET")
        secrets_after = self.adapter.list_secrets()
        assert not any(s.name == "GFO_TEST_SECRET" for s in secrets_after)

    # --- variable ---

    def test_51_variable_crud(self) -> None:
        """Variable の set → get → list → delete テスト。"""
        # クリーンアップ
        try:
            self.adapter.delete_variable("GFO_TEST_VAR")
        except Exception:
            pass
        var = self.adapter.set_variable("GFO_TEST_VAR", "test-value")
        assert var.name == "GFO_TEST_VAR"
        time.sleep(3)  # GitHub Actions variables API の反映ラグ対策
        got = self.adapter.get_variable("GFO_TEST_VAR")
        assert got.name == "GFO_TEST_VAR"
        assert got.value == "test-value"
        variables = self.adapter.list_variables()
        assert any(v.name == "GFO_TEST_VAR" for v in variables)
        self.adapter.delete_variable("GFO_TEST_VAR")
        time.sleep(3)
        variables_after = self.adapter.list_variables()
        assert not any(v.name == "GFO_TEST_VAR" for v in variables_after)



"""gfo ssh-key サブコマンドのハンドラ。"""

from __future__ import annotations

import argparse

from gfo.commands import get_adapter
from gfo.i18n import _
from gfo.output import output


def handle_view(args: argparse.Namespace, *, fmt: str, jq: str | None = None) -> None:
    """gfo ssh-key view <id> のハンドラ。"""
    adapter = get_adapter()
    key = adapter.get_ssh_key(args.id)
    output(key, fmt=fmt, jq=jq)


def handle_list(args: argparse.Namespace, *, fmt: str, jq: str | None = None) -> None:
    """gfo ssh-key list のハンドラ。"""
    adapter = get_adapter()
    keys = adapter.list_ssh_keys(limit=args.limit)
    output(keys, fmt=fmt, fields=["id", "title", "created_at"], jq=jq)


def handle_create(args: argparse.Namespace, *, fmt: str, jq: str | None = None) -> None:
    """gfo ssh-key create --title TITLE --key KEY のハンドラ。"""
    adapter = get_adapter()
    key = adapter.create_ssh_key(title=args.title, key=args.key)
    output(key, fmt=fmt, jq=jq)


def handle_delete(args: argparse.Namespace, *, fmt: str, jq: str | None = None) -> None:
    """gfo ssh-key delete <id> のハンドラ。"""
    adapter = get_adapter()
    adapter.delete_ssh_key(key_id=args.id)
    print(_("Deleted SSH key '{id}'.").format(id=args.id))



"""gfo.commands.variable のテスト。"""

from __future__ import annotations

import json
from unittest.mock import patch

import pytest

from gfo.adapter.base import Variable
from gfo.commands import variable as variable_cmd
from gfo.exceptions import GfoError, HttpError
from tests.test_commands.conftest import make_args, patch_adapter

SAMPLE_VAR = Variable(name="MY_VAR", value="val", created_at="2024-01-01", updated_at="2024-01-02")


class TestHandleList:
    def test_calls_list_variables(self, capsys):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.list_variables.return_value = [SAMPLE_VAR]
            args = make_args(limit=30, org=None)
            variable_cmd.handle_list(args, fmt="table")
        adapter.list_variables.assert_called_once_with(scope=None, limit=30)
        out = capsys.readouterr().out
        assert "MY_VAR" in out

    def test_json_format(self, capsys):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.list_variables.return_value = [SAMPLE_VAR]
            args = make_args(limit=30, org=None)
            variable_cmd.handle_list(args, fmt="json")
        out = capsys.readouterr().out
        parsed = json.loads(out)
        assert isinstance(parsed, list)
        assert parsed[0]["name"] == "MY_VAR"

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.list_variables.side_effect = HttpError(500, "Server error")
            args = make_args(limit=30, org=None)
            with pytest.raises(HttpError):
                variable_cmd.handle_list(args, fmt="table")


class TestHandleSet:
    def test_calls_set_variable(self):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.set_variable.return_value = SAMPLE_VAR
            args = make_args(name="MY_VAR", value="val", masked=False)
            variable_cmd.handle_set(args, fmt="table")
        adapter.set_variable.assert_called_once_with("MY_VAR", "val", scope=None, masked=False)

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.set_variable.side_effect = HttpError(403, "Forbidden")
            args = make_args(name="MY_VAR", value="val", masked=False)
            with pytest.raises(HttpError):
                variable_cmd.handle_set(args, fmt="table")


class TestHandleGet:
    def test_prints_value(self, capsys):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.get_variable.return_value = SAMPLE_VAR
            args = make_args(name="MY_VAR")
            variable_cmd.handle_get(args, fmt="table")
        out = capsys.readouterr().out
        assert out.strip() == "val"

    def test_json_format(self, capsys):
        """fmt="json" のとき output() 経由で JSON が出力される。"""
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.get_variable.return_value = SAMPLE_VAR
            args = make_args(name="MY_VAR")
            variable_cmd.handle_get(args, fmt="json")
        out = capsys.readouterr().out
        parsed = json.loads(out)
        assert parsed[0]["name"] == "MY_VAR"
        assert parsed[0]["value"] == "val"

    def test_jq_forces_json(self, capsys):
        """jq 引数指定時でも fmt="table" の場合 JSON 出力に切り替わる。"""
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.get_variable.return_value = SAMPLE_VAR
            args = make_args(name="MY_VAR")
            with patch("gfo.output.apply_jq_filter", return_value='"val"') as mock_jq:
                variable_cmd.handle_get(args, fmt="table", jq=".[0].value")
            mock_jq.assert_called_once()
            out = capsys.readouterr().out
            assert "val" in out

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.get_variable.side_effect = HttpError(404, "Not found")
            args = make_args(name="MY_VAR")
            with pytest.raises(HttpError):
                variable_cmd.handle_get(args, fmt="table")


class TestHandleDelete:
    def test_calls_delete(self):
        with patch_adapter("gfo.commands.variable") as adapter:
            args = make_args(name="MY_VAR")
            variable_cmd.handle_delete(args, fmt="table")
        adapter.delete_variable.assert_called_once_with("MY_VAR", scope=None)

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.delete_variable.side_effect = GfoError("not found")
            args = make_args(name="MY_VAR")
            with pytest.raises(GfoError, match="not found"):
                variable_cmd.handle_delete(args, fmt="table")


class TestOrgScope:
    def test_list_org_variables(self, capsys):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.list_variables.return_value = [SAMPLE_VAR]
            args = make_args(limit=30, org="my-org")
            variable_cmd.handle_list(args, fmt="table")
        adapter.list_variables.assert_called_once_with(scope="my-org", limit=30)

    def test_set_org_variable(self):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.set_variable.return_value = SAMPLE_VAR
            args = make_args(name="ORG_VAR", value="val", masked=False, org="my-org")
            variable_cmd.handle_set(args, fmt="table")
        adapter.set_variable.assert_called_once_with("ORG_VAR", "val", scope="my-org", masked=False)

    def test_delete_org_variable(self):
        with patch_adapter("gfo.commands.variable") as adapter:
            args = make_args(name="ORG_VAR", org="my-org")
            variable_cmd.handle_delete(args, fmt="table")
        adapter.delete_variable.assert_called_once_with("ORG_VAR", scope="my-org")


class TestHandleGetScope:
    def test_scope_passed_to_adapter(self, capsys):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.get_variable.return_value = Variable(
                name="FOO", value="bar", created_at="", updated_at=""
            )
            args = make_args(name="FOO", org="my-org")
            variable_cmd.handle_get(args, fmt="table")
        adapter.get_variable.assert_called_once_with("FOO", scope="my-org")

    def test_scope_none_by_default(self, capsys):
        with patch_adapter("gfo.commands.variable") as adapter:
            adapter.get_variable.return_value = Variable(
                name="FOO", value="bar", created_at="", updated_at=""
            )
            args = make_args(name="FOO")
            variable_cmd.handle_get(args, fmt="table")
        adapter.get_variable.assert_called_once_with("FOO", scope=None)



"""gfo.commands.ci のテスト。"""

from __future__ import annotations

from unittest.mock import patch

import pytest

from gfo.adapter.base import Artifact, Pipeline, Workflow
from gfo.commands import ci as ci_cmd
from gfo.exceptions import ConfigError, GfoError
from tests.test_commands.conftest import make_args, patch_adapter

SAMPLE_PIPELINE = Pipeline(
    id=123,
    status="success",
    ref="main",
    url="https://example.com/ci/123",
    created_at="2024-01-01T00:00:00Z",
)

SAMPLE_WORKFLOW = Workflow(
    id=1,
    name="CI",
    path=".github/workflows/ci.yml",
    state="active",
)

SAMPLE_ARTIFACT = Artifact(
    id=11,
    name="build-output",
    size=1024,
    url="https://example.com/artifacts/11",
    created_at="2024-01-01T00:00:00Z",
)


class TestHandleList:
    def test_calls_list_pipelines(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.list_pipelines.return_value = [SAMPLE_PIPELINE]
            args = make_args(ref=None, limit=30)
            ci_cmd.handle_list(args, fmt="table")
        adapter.list_pipelines.assert_called_once_with(ref=None, limit=30)

    def test_outputs_results(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.list_pipelines.return_value = [SAMPLE_PIPELINE]
            args = make_args(ref="main", limit=10)
            ci_cmd.handle_list(args, fmt="table")
        out = capsys.readouterr().out
        assert "success" in out


class TestHandleView:
    def test_calls_get_pipeline(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.get_pipeline.return_value = SAMPLE_PIPELINE
            args = make_args(id="123")
            ci_cmd.handle_view(args, fmt="table")
        adapter.get_pipeline.assert_called_once_with("123")


class TestHandleCancel:
    def test_calls_cancel_pipeline(self):
        with patch_adapter("gfo.commands.ci") as adapter:
            args = make_args(id="456")
            ci_cmd.handle_cancel(args, fmt="table")
        adapter.cancel_pipeline.assert_called_once_with("456")


class TestHandleTrigger:
    def test_calls_trigger_pipeline(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.trigger_pipeline.return_value = SAMPLE_PIPELINE
            args = make_args(ref="main", workflow="ci.yml", input=None)
            ci_cmd.handle_trigger(args, fmt="table")
        adapter.trigger_pipeline.assert_called_once_with("main", workflow="ci.yml", inputs=None)

    def test_json_output(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.trigger_pipeline.return_value = SAMPLE_PIPELINE
            args = make_args(ref="main", workflow="ci.yml", input=None)
            ci_cmd.handle_trigger(args, fmt="json")
        out = capsys.readouterr().out
        assert '"status"' in out

    def test_parses_inputs(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.trigger_pipeline.return_value = SAMPLE_PIPELINE
            args = make_args(ref="main", workflow="ci.yml", input=["FOO=bar", "BAZ=qux=1"])
            ci_cmd.handle_trigger(args, fmt="table")
        adapter.trigger_pipeline.assert_called_once_with(
            "main", workflow="ci.yml", inputs={"FOO": "bar", "BAZ": "qux=1"}
        )

    def test_invalid_input_format(self):
        with patch_adapter("gfo.commands.ci"):
            args = make_args(ref="main", workflow="ci.yml", input=["INVALID"])
            with pytest.raises(ConfigError, match="Invalid input format"):
                ci_cmd.handle_trigger(args, fmt="table")

    def test_no_workflow(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.trigger_pipeline.return_value = SAMPLE_PIPELINE
            args = make_args(ref="main", workflow=None, input=None)
            ci_cmd.handle_trigger(args, fmt="table")
        adapter.trigger_pipeline.assert_called_once_with("main", workflow=None, inputs=None)

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.trigger_pipeline.side_effect = GfoError("trigger failed")
            args = make_args(ref="main", workflow="ci.yml", input=None)
            with pytest.raises(GfoError, match="trigger failed"):
                ci_cmd.handle_trigger(args, fmt="table")


class TestHandleRetry:
    def test_calls_retry_pipeline(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.retry_pipeline.return_value = SAMPLE_PIPELINE
            args = make_args(id="123")
            ci_cmd.handle_retry(args, fmt="table")
        adapter.retry_pipeline.assert_called_once_with("123")

    def test_json_output(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.retry_pipeline.return_value = SAMPLE_PIPELINE
            args = make_args(id="123")
            ci_cmd.handle_retry(args, fmt="json")
        out = capsys.readouterr().out
        assert '"status"' in out

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.retry_pipeline.side_effect = GfoError("retry failed")
            args = make_args(id="123")
            with pytest.raises(GfoError, match="retry failed"):
                ci_cmd.handle_retry(args, fmt="table")


class TestHandleLogs:
    def test_calls_get_pipeline_logs(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.get_pipeline_logs.return_value = "log output here"
            args = make_args(id="123", job=None)
            ci_cmd.handle_logs(args, fmt="table")
        adapter.get_pipeline_logs.assert_called_once_with("123", job_id=None)
        out = capsys.readouterr().out
        assert "log output here" in out

    def test_with_job_id(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.get_pipeline_logs.return_value = "job log output"
            args = make_args(id="123", job="456")
            ci_cmd.handle_logs(args, fmt="table")
        adapter.get_pipeline_logs.assert_called_once_with("123", job_id="456")
        out = capsys.readouterr().out
        assert "job log output" in out

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.get_pipeline_logs.side_effect = GfoError("logs failed")
            args = make_args(id="123", job=None)
            with pytest.raises(GfoError, match="logs failed"):
                ci_cmd.handle_logs(args, fmt="table")


class TestHandleWatch:
    def test_watch_immediate_success(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.get_pipeline.return_value = SAMPLE_PIPELINE
            args = make_args(id="123", interval=1)
            ci_cmd.handle_watch(args, fmt="json")
        adapter.get_pipeline.assert_called_once_with("123")
        out = capsys.readouterr().out
        assert '"status"' in out

    def test_watch_polls_until_done(self, capsys):
        running_pipeline = Pipeline(
            id=123,
            status="running",
            ref="main",
            url="https://example.com/ci/123",
            created_at="2024-01-01T00:00:00Z",
        )
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.get_pipeline.side_effect = [running_pipeline, SAMPLE_PIPELINE]
            args = make_args(id="123", interval=0)
            with patch("time.sleep") as mock_sleep:
                ci_cmd.handle_watch(args, fmt="table")
        mock_sleep.assert_called()
        assert adapter.get_pipeline.call_count == 2

    def test_watch_failure(self, capsys):
        failed_pipeline = Pipeline(
            id=123,
            status="failure",
            ref="main",
            url="https://example.com/ci/123",
            created_at="2024-01-01T00:00:00Z",
        )
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.get_pipeline.return_value = failed_pipeline
            args = make_args(id="123", interval=1)
            ci_cmd.handle_watch(args, fmt="json")
        out = capsys.readouterr().out
        assert '"failure"' in out

    def test_watch_timeout(self, capsys):
        running_pipeline = Pipeline(
            id=123,
            status="running",
            ref="main",
            url="https://example.com/ci/123",
            created_at="2024-01-01T00:00:00Z",
        )
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.get_pipeline.return_value = running_pipeline
            args = make_args(id="123", interval=0, timeout=1)
            with patch("time.sleep"), patch("time.monotonic", side_effect=[0, 0, 2]):
                ci_cmd.handle_watch(args, fmt="table")
        out = capsys.readouterr().out
        assert "Timed out" in out

    def test_watch_error_propagation(self):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.get_pipeline.side_effect = GfoError("watch failed")
            args = make_args(id="123", interval=1)
            with pytest.raises(GfoError, match="watch failed"):
                ci_cmd.handle_watch(args, fmt="table")


class TestHandleDownload:
    def test_calls_download_run_logs(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.download_run_logs.return_value = "/tmp/logs-123.zip"
            args = make_args(id="123", job=None, dir=".")
            ci_cmd.handle_download(args, fmt="table")
        adapter.download_run_logs.assert_called_once_with("123", job_id=None, output_dir=".")
        out = capsys.readouterr().out
        assert "logs-123.zip" in out

    def test_with_job_id(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.download_run_logs.return_value = "/tmp/logs-123-job-42.txt"
            args = make_args(id="123", job="42", dir="/tmp")
            ci_cmd.handle_download(args, fmt="table")
        adapter.download_run_logs.assert_called_once_with("123", job_id="42", output_dir="/tmp")

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.download_run_logs.side_effect = GfoError("download failed")
            args = make_args(id="123", job=None, dir=".")
            with pytest.raises(GfoError, match="download failed"):
                ci_cmd.handle_download(args, fmt="table")


class TestHandleWorkflow:
    def test_list(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.list_workflows.return_value = [SAMPLE_WORKFLOW]
            args = make_args(workflow_action="list", limit=30)
            ci_cmd.handle_workflow(args, fmt="table")
        adapter.list_workflows.assert_called_once_with(limit=30)
        out = capsys.readouterr().out
        assert "CI" in out

    def test_list_json(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.list_workflows.return_value = [SAMPLE_WORKFLOW]
            args = make_args(workflow_action="list", limit=30)
            ci_cmd.handle_workflow(args, fmt="json")
        out = capsys.readouterr().out
        assert '"name"' in out

    def test_enable(self):
        with patch_adapter("gfo.commands.ci") as adapter:
            args = make_args(workflow_action="enable", id="1")
            ci_cmd.handle_workflow(args, fmt="table")
        adapter.enable_workflow.assert_called_once_with("1")

    def test_disable(self):
        with patch_adapter("gfo.commands.ci") as adapter:
            args = make_args(workflow_action="disable", id="1")
            ci_cmd.handle_workflow(args, fmt="table")
        adapter.disable_workflow.assert_called_once_with("1")

    def test_enable_prints_message(self, capsys):
        with patch_adapter("gfo.commands.ci"):
            args = make_args(workflow_action="enable", id="42")
            ci_cmd.handle_workflow(args, fmt="table")
        out = capsys.readouterr().out
        assert "42" in out

    def test_disable_prints_message(self, capsys):
        with patch_adapter("gfo.commands.ci"):
            args = make_args(workflow_action="disable", id="42")
            ci_cmd.handle_workflow(args, fmt="table")
        out = capsys.readouterr().out
        assert "42" in out

    def test_no_action(self):
        with patch_adapter("gfo.commands.ci"):
            args = make_args(workflow_action=None)
            with pytest.raises(ConfigError):
                ci_cmd.handle_workflow(args, fmt="table")

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.list_workflows.side_effect = GfoError("workflow failed")
            args = make_args(workflow_action="list", limit=30)
            with pytest.raises(GfoError, match="workflow failed"):
                ci_cmd.handle_workflow(args, fmt="table")


class TestHandleArtifact:
    def test_list(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.list_artifacts.return_value = [SAMPLE_ARTIFACT]
            args = make_args(artifact_action="list", run_id="300", limit=30)
            ci_cmd.handle_artifact(args, fmt="table")
        adapter.list_artifacts.assert_called_once_with("300", limit=30)
        out = capsys.readouterr().out
        assert "build-output" in out

    def test_list_json(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.list_artifacts.return_value = [SAMPLE_ARTIFACT]
            args = make_args(artifact_action="list", run_id="300", limit=30)
            ci_cmd.handle_artifact(args, fmt="json")
        out = capsys.readouterr().out
        assert '"name"' in out

    def test_download(self, capsys):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.download_artifact.return_value = "/tmp/build-output.zip"
            args = make_args(artifact_action="download", run_id="300", artifact_id="11", dir=".")
            ci_cmd.handle_artifact(args, fmt="table")
        adapter.download_artifact.assert_called_once_with("300", "11", output_dir=".")
        out = capsys.readouterr().out
        assert "build-output.zip" in out

    def test_no_action(self):
        with patch_adapter("gfo.commands.ci"):
            args = make_args(artifact_action=None)
            with pytest.raises(ConfigError):
                ci_cmd.handle_artifact(args, fmt="table")

    def test_error_propagation(self):
        with patch_adapter("gfo.commands.ci") as adapter:
            adapter.list_artifacts.side_effect = GfoError("artifact failed")
            args = make_args(artifact_action="list", run_id="300", limit=30)
            with pytest.raises(GfoError, match="artifact failed"):
                ci_cmd.handle_artifact(args, fmt="table")
